r"""
Analytic Acquisition Functions for Multi-objective Bayesian optimization.

References

.. [Yang2019]
    Yang, K., Emmerich, M., Deutz, A. et al. Efficient computation of expected
    hypervolume improvement using box decomposition algorithms. J Glob Optim 75,
    3–34 (2019)

"""


from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from itertools import product
from typing import Any, List, Optional

import numpy as np
import torch
from torch import Tensor
from torch.distributions import Normal
from torch.nn import Module

from ..extra.multi_objective.box_decompositions.non_dominated import NondominatedPartitioning


class AcquisitionFunction(Module, ABC):
    r"""Abstract base class for acquisition functions."""

    def __init__(self, model) -> None:
        r"""Constructor for the AcquisitionFunction base class.

        Args:
            model: A fitted model.
        """
        super().__init__()
        self.model = model

    def set_X_pending(self, X_pending: Optional[Tensor] = None) -> None:
        r"""Informs the acquisition function about pending design points.

        Args:
            X_pending: `n x d` Tensor with `n` `d`-dim design points that have
                been submitted for evaluation but have not yet been evaluated.
        """
        if X_pending is not None:
            if X_pending.requires_grad:
                warnings.warn(
                    "Pending points require a gradient but the acquisition function"
                    " will not provide a gradient to these points."
                )
            self.X_pending = X_pending.detach().clone()
        else:
            self.X_pending = X_pending

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the acquisition function on the candidate set X.

        Args:
            X: A `(b) x q x d`-dim Tensor of `(b)` t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `(b)`-dim Tensor of acquisition function values at the given
            design points `X`.
        """
        ...


class MultiObjectiveAnalyticAcquisitionFunction(AcquisitionFunction):
    r"""Abstract base class for Multi-Objective batch acquisition functions."""

    def __init__(self, model, objective: Any = None) -> None:
        r"""Constructor for the MultiObjectiveAnalyticAcquisitionFunction base class.

        Args:
            model: A fitted model.
            objective: An AnalyticMultiOutputObjective (optional).
        """
        super().__init__(model=model)
        self.objective = objective

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Takes in a `batch_shape x 1 x d` X Tensor of t-batches with `1` `d`-dim
        design point each, and returns a Tensor with shape `batch_shape'`, where
        `batch_shape'` is the broadcasted batch shape of model and input `X`.
        """
        ...

    def set_X_pending(self, X_pending: Optional[Tensor] = None) -> None:
        raise ValueError("Analytic acquisition functions do not account for X_pending yet.")


class EHVI(MultiObjectiveAnalyticAcquisitionFunction):
    """Expected Hypervolume Improvement"""

    def __init__(
        self,
        model,
        ref_point: List[float],
        partitioning: NondominatedPartitioning,
    ) -> None:
        r"""Expected Hypervolume Improvement supporting m>=2 outcomes.

        This implements the computes EHVI using the algorithm from [Yang2019]_, but
        additionally computes gradients via auto-differentiation as proposed by
        [Daulton2020qehvi]_.

        Note: this is currently inefficient in two ways due to the binary partitioning
        algorithm that we use for the box decomposition:

            - We have more boxes in our decomposition
            - If we used a box decomposition that used `inf` as the upper bound for
                the last dimension *in all hypercells*, then we could reduce the number
                of terms we need to compute from 2^m to 2^(m-1). [Yang2019]_ do this
                by using DKLV17 and LKF17 for the box decomposition.

        TODO: Use DKLV17 and LKF17 for the box decomposition as in [Yang2019]_ for
        greater efficiency.

        TODO: Add support for outcome constraints.

        Example:
            >>> model = SingleTaskGP(train_X, train_Y)
            >>> ref_point = [0.0, 0.0]
            >>> EHVI = ExpectedHypervolumeImprovement(model, ref_point, partitioning)
            >>> ehvi = EHVI(test_X)

        Args:
            model: A fitted model.
            ref_point: A list with `m` elements representing the reference point (in the
                outcome space) w.r.t. to which compute the hypervolume. This is a
                reference point for the objective values (i.e. after applying
                `objective` to the samples).
            partitioning: A `NondominatedPartitioning` module that provides the non-
                dominated front and a partitioning of the non-dominated space in hyper-
                rectangles.
            objective: An `AnalyticMultiOutputObjective`.
        """
        # TODO: we could refactor this __init__ logic into a
        # HypervolumeAcquisitionFunction Mixin
        if len(ref_point) != partitioning.num_outcomes:
            raise ValueError(
                "The length of the reference point must match the number of outcomes. "
                f"Got ref_point with {len(ref_point)} elements, but expected "
                f"{partitioning.num_outcomes}."
            )
        ref_point = torch.tensor(
            ref_point,
            dtype=partitioning.pareto_Y.dtype,
            device=partitioning.pareto_Y.device,
        )
        better_than_ref = (partitioning.pareto_Y > ref_point).all(dim=1)
        if not better_than_ref.any() and partitioning.pareto_Y.shape[0] > 0:
            raise ValueError("At least one pareto point must be better than the reference point.")

        super().__init__(model=model)
        self.register_buffer("ref_point", ref_point)
        self.partitioning = partitioning
        cell_bounds = self.partitioning.get_hypercell_bounds()
        self.register_buffer("cell_lower_bounds", cell_bounds[0])
        self.register_buffer("cell_upper_bounds", cell_bounds[1])
        # create indexing tensor of shape `2^m x m`
        self._cross_product_indices = torch.tensor(
            list(product(*[[0, 1] for _ in range(ref_point.shape[0])])),
            dtype=torch.long,
            device=ref_point.device,
        )
        self.normal = Normal(0, 1)

    def psi(self, lower: Tensor, upper: Tensor, mu: Tensor, sigma: Tensor) -> None:
        r"""Compute Psi function.

        For each cell i and outcome k:

            Psi(lower_{i,k}, upper_{i,k}, mu_k, sigma_k) = (
            sigma_k * PDF((upper_{i,k} - mu_k) / sigma_k) + (
            mu_k - lower_{i,k}
            ) * (1 - CDF(upper_{i,k} - mu_k) / sigma_k)
            )

        See Equation 19 in [Yang2019]_ for more details.

        Args:
            lower: A `num_cells x m`-dim tensor of lower cell bounds
            upper: A `num_cells x m`-dim tensor of upper cell bounds
            mu: A `batch_shape x 1 x m`-dim tensor of means
            sigma: A `batch_shape x 1 x m`-dim tensor of standard deviations (clamped).

        Returns:
            A `batch_shape x num_cells x m`-dim tensor of values.
        """
        u = (upper - mu) / sigma
        return sigma * self.normal.log_prob(u).exp() + (mu - lower) * (1 - self.normal.cdf(u))

    def nu(self, lower: Tensor, upper: Tensor, mu: Tensor, sigma: Tensor) -> None:
        r"""Compute Nu function.

        For each cell i and outcome k:

            nu(lower_{i,k}, upper_{i,k}, mu_k, sigma_k) = (
            upper_{i,k} - lower_{i,k}
            ) * (1 - CDF((upper_{i,k} - mu_k) / sigma_k))

        See Equation 25 in [Yang2019]_ for more details.

        Args:
            lower: A `num_cells x m`-dim tensor of lower cell bounds
            upper: A `num_cells x m`-dim tensor of upper cell bounds
            mu: A `batch_shape x 1 x m`-dim tensor of means
            sigma: A `batch_shape x 1 x m`-dim tensor of standard deviations (clamped).

        Returns:
            A `batch_shape x num_cells x m`-dim tensor of values.
        """
        return (upper - lower) * (1 - self.normal.cdf((upper - mu) / sigma))

    def forward(self, X: np.ndarray) -> Tensor:
        # convert the input if its shape it not correct
        if isinstance(X, list):
            X = np.array(X, dtype=object)
        if len(X.shape) == 1:
            X = X.reshape(1, -1)
        # get the mean and variance
        mean, MSE = self.model.predict(X, eval_MSE=True)
        mean = torch.Tensor(mean).unsqueeze(0)
        variance = torch.Tensor(MSE).unsqueeze(0)
        mu, sigma = mean, variance.clamp_min(1e-9).sqrt()
        # clamp here, since upper_bounds will contain `inf`s, which
        # are not differentiable
        cell_upper_bounds = self.cell_upper_bounds.clamp_max(
            1e10 if X.dtype == np.float64 else 1e8
        )
        # Compute psi(lower_i, upper_i, mu_i, sigma_i) for i=0, ... m-2
        psi_lu = self.psi(
            lower=self.cell_lower_bounds, upper=cell_upper_bounds, mu=mu, sigma=sigma
        )
        # Compute psi(lower_m, lower_m, mu_m, sigma_m)
        psi_ll = self.psi(
            lower=self.cell_lower_bounds,
            upper=self.cell_lower_bounds,
            mu=mu,
            sigma=sigma,
        )
        # Compute nu(lower_m, upper_m, mu_m, sigma_m)
        nu = self.nu(lower=self.cell_lower_bounds, upper=cell_upper_bounds, mu=mu, sigma=sigma)
        # compute the difference psi_ll - psi_lu
        psi_diff = psi_ll - psi_lu

        # this is batch_shape x num_cells x 2 x (m-1)
        stacked_factors = torch.stack([psi_diff, nu], dim=-2)

        # Take the cross product of psi_diff and nu across all outcomes
        # e.g. for m = 2
        # for each batch and cell, compute
        # [psi_diff_0, psi_diff_1]
        # [nu_0, psi_diff_1]
        # [psi_diff_0, nu_1]
        # [nu_0, nu_1]
        # this tensor has shape: `batch_shape x num_cells x 2^m x m`
        all_factors_up_to_last = stacked_factors.gather(
            dim=-2,
            index=self._cross_product_indices.expand(
                stacked_factors.shape[:-2] + self._cross_product_indices.shape
            ),
        )
        # compute product for all 2^m terms,
        # sum across all terms and hypercells
        return all_factors_up_to_last.prod(dim=-1).sum(dim=-1).sum(dim=-1).cpu().numpy()



import os
import shutil
import subprocess
import time

import numpy as np
import requests

data = {
    "search_param": {
        "emissivity": {"type": "r", "range": [0.95, 1], "N": 2, "precision": 2},
        "offset": {"type": "r", "range": [-10, 10], "N": 2, "precision": 2},
        "power": {"type": "r", "range": [3.2, 3.8], "N": 1, "precision": 2},  # 数值精度：小数点后位数
    },
    "bo_param": {
        "n_job": 1,  # 服务器上并行进程数
        "n_point": 1,  # 每次迭代返回参数值个数
        "max_iter": 20,  # 最大迭代次数
        "DoE_size": 3,  # 初始（第一代）采样点个数，其一般与`n_point`相等
        "minimize": True,  # 最大化/最小化
        "n_obj": 1,
    },
}

address = "http://127.0.0.1:7200"


def obj_func_dict_eval2(par):
    """范例目标函数，其输入`par`为一个包含了一组候选参数的字典
    这些参数在字典中的键值是`search_space`中给出的参数名，如'continuous'，'nominal'
    这个函数应该返回由用户自定义的目标值。在这里我们给出了一个很简单的范例，其不蕴含任何实际意义
    """
    x_r = np.asarray([par[k] for k in par.keys() if k.startswith("emissivity")])
    x_r2 = np.asarray([par[k] for k in par.keys() if k.startswith("offset")])
    x_i = par["power"]
    return (
        np.sum(x_r ** 2.0)
        + abs(x_i - 3.5)
        + np.sum(x_r2 ** 2.0)
        + np.random.randn() * np.sqrt(0.5)
    )


env = os.environ.copy()
if not "PYTHONPATH" in env:
    env["PYTHONPATH"] = ""

env["PYTHONPATH"] = "../" + env["PYTHONPATH"]

# 我们先启动优化器服务进程
proc = subprocess.Popen(
    ["python3", "-m", "bayes_optim.simple_http_server", "-w", "7200", "-v"], env=env
)
time.sleep(3)

# 请求创建新的优化任务，其初始化数据由一个json数据`data`给定
r = requests.post(address, json=data)
job_id = r.json()["job_id"]
print("Job id is %s" % (job_id))

for i in range(3):
    print("iteration %d" % (i))

    # 请求一组候选参数值用于测试。请求时必须附加之前初始化时返回的任务`id`
    r = requests.get(address, params={"ask": "null", "job_id": job_id})
    tell_data = r.json()

    # 运行并评估优化服务器返回的候选参数
    y = [obj_func_dict_eval2(_) for _ in tell_data["X"]]
    tell_data["y"] = y

    print(tell_data)
    r = requests.post(address, json=tell_data)

# 在优化任务结束后，我们向服务器发出请求结束掉这个任务并清除服务器上的相关文件
r = requests.get(address, params={"finalize": "null", "job_id": job_id})

proc.kill()
shutil.rmtree("7200")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test function interface.

This module implements class LoggingFunction which is the main class for
logging a whole experiment.

Example:

    >>> import fgeneric, bbobbenchmarks as bb
    >>> f = fgeneric.LoggingFunction('fgeneric_doctest_no_algorihm', 'no-name-algorithm')  # like a LoggingFunction class
    >>> f.setfun(bb.F1()).ftarget  # assumes that F1().fopt is accessible, prints target value
    -92.649999990000012
    >>> f([1,2,3])  # call f1 while logging the results, same as f.evalfun(x)
    -61.233399680000005
    >>> f.finalizerun()  # reset counter for f-evaluations etc

For other examples of how to use fgeneric, see exampleexperiment.py and exampletiming.py

Test:

    python fgeneric.py

runs doctest, using -v option gives verbosity.

"""

# Changes:
# 12/03/08: renamed cocoexp.Logger to fgeneric.LoggingFunction and setfun returns self
# 12/02/22: def setfun, the second argument has become optional and attribute
#           fopt of the first argument is used, if the second is not given

import sys
import os
import errno
import warnings
from pdb import set_trace
import copy
import numpy as np

deltaftarget = 1e-8
nb_evaluations_always_written = (
    "1"  # '100 + 10 * dim'  # 100 + dim;10*dim add about 7;17MB to final data
)
nb_triggers_per_delta_f_decade = 5  # using 10 should be perfectly fine
fileprefix = "bbobexp"


class LoggingFunction(object):
    """Class for a function that records data from experiments with a given
       algorithm and parameter settings.

    This class provides recording facilities:

    * index files (:file:`info` extension) record information on an
      experiment
    * data files (:file:`dat`, :file:`tdat` extensions) which record
      number of function evaluations and function values for consecutive
      runs.

    Once an instance of this class is provided with a test function and
    the function optimal value, the instance or its method :meth:`evalfun`
    can be called to evaluate the function.

    At the end of the run, :meth:`finalizerun` must be called.

    Either the provided test function is an object with an evaluation
    method :meth:`_evalfull` which returns noisy and noiseless values
    (tuple of length 2). Otherwise, the test function should be
    callable and return a scalar if called with tuple (0., 0.).

    """

    nbptsevals = 20.0  # number of trigger per decade of function evaluations
    nbptsf = float(
        nb_triggers_per_delta_f_decade
    )  # number of trigger per decade of function values, float to prevent integer division

    def __call__(self, x, *args, **kwargs):  # makes the instances callable
        """Returns objective function value"""
        return self.evalfun(x, *args, **kwargs)

    class __Eval(object):
        """Class for the object lasteval."""

        def __init__(self):
            # TODO: check what is actually needed here.
            self.num = 0.0
            self.f = np.inf  # minimization...
            self.bestf = np.inf  # minimization...
            self.fnoisy = np.inf  # minimization...
            self.bestfnoisy = np.inf  # minimization...
            self.is_written = False

        def sprintData(self, fopt):
            """Format data for printing."""

            res = "%d %+10.9e %+10.9e %+10.9e %+10.9e" % (
                self.num,
                self.f - fopt,
                self.bestf - fopt,
                self.fnoisy,
                self.bestfnoisy,
            )

            if len(self.x) < 22:
                tmp = []
                for i in self.x:
                    tmp.append(" %+5.4e" % i)
                res += "".join(tmp)
            res += "\n"
            return res

        def update(self, fnoisy, f, x):
            """Update the content of lasteval."""
            try:
                self.num += len(f)
                self.f = f[-1]
                self.fnoisy = fnoisy[-1]
                self.x = x[-1]
            except TypeError:
                self.num += 1
                self.f = f
                self.fnoisy = fnoisy
                self.x = x
            bestf = np.min(f)
            if bestf < self.bestf:
                self.bestf = bestf
            bestfnoisy = np.min(fnoisy)
            if bestfnoisy < self.bestfnoisy:
                self.bestfnoisy = bestfnoisy
            self.is_written = False

    def __init__(self, datapath, algid="not-specified", comments="", inputformat="row"):
        """Initialize LoggingFunction for an experiment. Before the
        LoggingFunction can be used as an objective
        function, method :meth:`setfun` must be called.

        :param string datapath: Output folder name
        :param string algid: name of algorithm
        :param string comments: complementary information on experiment
          (parameter settings and such)
        :param string inputformat: 'row' (default) or 'col', determines
          the shape of the input data.

        """
        self.initialize(datapath, algid, comments, inputformat)

    def initialize(self, datapath, algid="not-specified", comments="", inputformat="row"):
        """Initialize LoggingFunction with a data path and further
        infos. Before the LoggingFunction can be used as an objective
        function, method :meth:`setfun` must be called.

        """
        self.datapath = datapath
        self.algid = algid
        self.comments = comments
        self.inputformat = inputformat
        self.fileprefix = fileprefix
        self._is_setfun = False
        self._is_setdim = False
        self._is_finalized = True
        self._is_samefun = False
        self._is_samedim = False

    def __del__(self):
        """Destructor.

        Will attempt to finalize a run when the current instance is
        about to be deleted.

        """
        try:
            if not self._is_finalized:
                self.finalizerun()
        except AttributeError:
            pass

    def _is_ready(self):
        res = self._is_setdim and self._is_setfun and not self._is_finalized
        return res

    def _readytostart(self):
        # index entry, data files
        filename = "%s_f%s_DIM%d" % (self.datafileprefix, str(self.funId), self._dim)

        res = []
        if not (self._is_samefun and self._is_samedim) or not os.path.exists(self.indexfile):
            i = 0
            while os.path.exists(filename + ".dat"):
                i += 1
                filename = "%s-%02d_f%s_DIM%d" % (
                    self.datafileprefix,
                    i,
                    str(self.funId),
                    self._dim,
                )
            self.datafile = filename + ".tdat"
            self.hdatafile = filename + ".dat"
            self.rdatafile = filename + ".rdat"

            if os.path.exists(self.indexfile):
                res.append("\n")  # was: res += '\n'

            if isinstance(self.funId, str):
                tmp = "'%s'" % self.funId
            else:
                tmp = str(self.funId)
            res.append("funcId = %s" % tmp)
            for i in self._fun_kwargs.items():
                if isinstance(i[1], str):
                    tmp = "'%s'" % i[1]
                else:
                    tmp = str(i[1])
                res.append(", %s = %s" % (str(i[0]), i[1]))
            res.append(
                ", DIM = %d, Precision = %.3e, algId = '%s'\n"
                % (self._dim, self.precision, self.algid)
            )
            res.append(
                "%% %s\n%s" % (self.comments, os.path.relpath(self.hdatafile, self.datapath))
            )

        if isinstance(self.iinstance, str):
            tmp = "'%s'" % self.iinstance
        else:
            tmp = str(self.iinstance)
        res.append(", %s" % tmp)
        f = open(self.indexfile, "a")
        f.writelines(res)  # does not add line separator
        f.close()
        self._is_finalized = False  # Just starting.

        for datafile in (self.datafile, self.hdatafile, self.rdatafile):
            filepath, filename = os.path.split(datafile)
            try:
                os.makedirs(filepath)
            except OSError as xxx_todo_changeme:
                (err, strerror) = xxx_todo_changeme.args
                if err in (errno.EEXIST, 183):  # 183 it is on Windows 7
                    pass
                else:
                    print(errno, strerror)
            f = open(datafile, "a")
            f.write(
                "%% function evaluation | noise-free fitness - Fopt"
                " (%13.12e) | best noise-free fitness - Fopt | measured "
                "fitness | best measured fitness | x1 | x2...\n" % self.fopt
            )
            f.close()

    def evalfun(self, inputx, *args, **kwargs):
        """Evaluate the function, return objective function value.

        Positional and keyword arguments args and kwargs are directly
        passed to the test function evaluation method.

        """
        # This block is the opposite in Matlab!
        if self._is_rowformat:
            x = np.asarray(inputx)
        else:
            x = np.transpose(inputx)

        curshape = np.shape(x)
        dim = curshape[-1]
        if len(curshape) < 2:
            popsi = 1
        else:
            popsi = curshape[0]

        if not self._is_setdim or self._dim != dim:
            self._setdim(dim)

        if not self._is_ready():
            self._readytostart()

        out = self._fun_evalfull(x, *args, **kwargs)
        try:
            fvalue, ftrue = out
        except TypeError:
            fvalue = out
            ftrue = out
            self._fun_eval = self._fun_evalfull
            self._fun_evalfull = lambda x: tuple([self._fun_eval(x)] * 2)

        if (
            self.lasteval.num + popsi >= self.evalsTrigger
            or np.min(ftrue) - self.fopt < self.fTrigger
        ):  # need to write something
            buffr = []
            hbuffr = []
            for j in range(0, popsi):
                try:
                    fvaluej = fvalue[j]
                    ftruej = ftrue[j]
                    xj = x[j]
                except (IndexError, ValueError, TypeError):  # cannot slice a 0-d array
                    fvaluej = fvalue
                    ftruej = ftrue
                    xj = x
                self.lasteval.update(fvaluej, ftruej, xj)

                if self.lasteval.num >= self.evalsTrigger:
                    buffr.append(self.lasteval.sprintData(self.fopt))
                    while self.lasteval.num >= np.floor(
                        10 ** (self.idxEvalsTrigger / self.nbptsevals)
                    ):
                        self.idxEvalsTrigger += 1
                    while self.lasteval.num >= dim * 10 ** self.idxDIMEvalsTrigger:
                        self.idxDIMEvalsTrigger += 1
                    self.evalsTrigger = min(
                        np.floor(10 ** (self.idxEvalsTrigger / self.nbptsevals)),
                        dim * 10 ** self.idxDIMEvalsTrigger,
                    )
                    if self.lasteval.num < self.nbFirstEvalsToAlwaysWrite:
                        self.evalsTrigger = self.lasteval.num + 1
                    self.lasteval.is_written = True

                if ftruej - self.fopt < self.fTrigger:  # minimization
                    hbuffr.append(self.lasteval.sprintData(self.fopt))
                    if ftruej <= self.fopt:
                        self.fTrigger = -np.inf
                    else:
                        if np.isinf(self.idxFTrigger):
                            self.idxFTrigger = np.ceil(np.log10(ftruej - self.fopt)) * self.nbptsf
                        while ftruej - self.fopt <= 10 ** (self.idxFTrigger / self.nbptsf):
                            self.idxFTrigger -= 1
                        self.fTrigger = min(
                            self.fTrigger, 10 ** (self.idxFTrigger / self.nbptsf)
                        )  # TODO: why?
            # write
            if buffr:
                f = open(self.datafile, "a")
                f.writelines(buffr)
                f.close()
            if hbuffr:
                f = open(self.hdatafile, "a")
                f.writelines(hbuffr)
                f.close()
        else:
            self.lasteval.update(fvalue, ftrue, x)

        return fvalue

    def finalizerun(self):
        """Write the last bit of information for a given run.

        Calling this method at the end of a run is necessary as some
        information are not written otherwise.

        """
        if self._is_finalized:
            warnings.warn("Run was never started.")
            return
        if not self.lasteval.is_written:
            if not os.path.exists(self.datafile):
                warnings.warn(
                    "The data file %s is not found. "
                    "Data will be appended to an empty file. Previously "
                    "obtained data may be missing." % self.datafile
                )
            f = open(self.datafile, "a")
            f.write(self.lasteval.sprintData(self.fopt))
            f.close()

        # write in self.indexfile
        if not os.path.exists(self.indexfile):
            warnings.warn(
                "The index file %s is not found. "
                "Data will be appended to an empty file. Previously "
                "obtained data may be missing." % self.indexfile
            )
        f = open(self.indexfile, "a")
        f.write(":%d|%.1e" % (self.lasteval.num, self.lasteval.bestf - self.fopt - self.precision))
        f.close()
        self._is_finalized = True

    # GETTERS/SETTERS
    def _getdatapath(self):
        """Main path for storing the data."""
        return self._datapath

    def _setdatapath(self, datapath):
        try:
            os.makedirs(datapath)
        except AttributeError:
            print("Input argument datapath is an invalid datapath.", file=sys.stderr)
            raise
        except OSError as xxx_todo_changeme1:
            (err, strerror) = xxx_todo_changeme1.args
            if err in (errno.EEXIST, 183):  # 183 it is on Windows 7
                pass
            else:
                print(errno, strerror)
        self._datapath = datapath

    datapath = property(_getdatapath, _setdatapath)

    def _getalgid(self):
        """String representing the tested algorithm."""
        return self._algid

    def _setalgid(self, algid):
        self._algid = str(algid)

    algid = property(_getalgid, _setalgid)

    def _getcomments(self):
        """String with any information on the experiment set.

        It is recommended to put all of the parameter settings here.
        """
        return self._comments

    def _setcomments(self, comments):
        self._comments = str(comments)

    comments = property(_getcomments, _setcomments)

    def _getinputformat(self):
        """String giving the input format, either 'row' (default) or 'col'."""
        return self._inputformat

    def _setinputformat(self, inputformat):
        if inputformat in ("row", "col"):
            self._inputformat = inputformat
        else:
            warnings.warn(
                "The inputFormat input argument is expected to"
                + " match either 'col' or 'row'. Attempting to"
                + "use default ('row')."
            )
            self._inputformat = "row"

        self._is_rowformat = self._inputformat == "row"

    inputformat = property(_getinputformat, _setinputformat)

    def _getfbest(self):
        """Returns the best function value obtained."""

        return self.lasteval.bestf

    fbest = property(_getfbest)
    best = property(_getfbest)

    def _getftarget(self):
        """Returns the target function value."""

        try:
            return self.fopt + self.precision
        except AttributeError:
            raise  # TODO

    ftarget = property(_getftarget)

    def _getevaluations(self):
        """Number of function evaluations so far."""
        if not hasattr(self, "lasteval") or not self._is_setdim:
            # when self._is_setfun and not self._is_setdim,
            # self.lasteval.num might not be 0. TODO: prevent this inconsistency.
            return 0.0  # should be synchronized with what is assigned to __Eval.num
        else:
            return self.lasteval.num

    evaluations = property(_getevaluations)

    def _getfun(self):
        """Current test function to be evaluated."""
        return self._fun

    def setfun(
        self,
        fun,
        fopt=None,
        funId="undefined",
        iinstance="undefined",
        dftarget=deltaftarget,
        **kwargs,
    ):
        """Set test function, returns an evaluation method and a target

        Using this method is necessary to start the creation of a log of
        an experiment.

        :param fun: test function (could be the evaluation method or an
                    instance of bbobbenchmarks.)
                    has to return a tuple of length 2: a noisy and a
                    noiseless value. If the function is noiseless, you
                    can provide (lambda x: tuple([f(x)] * 2)).
        :param float fopt: optimum value of the test function, used to define
                           the f-values to be recorded
        :param float dftarget: target precision, only used to set the ftarget
            property attribute (must only be used for checking for final termination)
        :param funId: function identifier, typically a number, funId must have
                    a string representation and is used for file names (avoid
                    special characters)
        :param iinstance: instance of the function (has to have string
                          representation)
        :param kwargs: additional descriptions

        :returns: ``self``, ``self`` can be called as an objective function,
                the target function value is self.ftarget = fopt + dftarget

        """
        if not self._is_finalized:
            self.finalizerun()

        self._is_setdim = False

        if fopt is None:
            fopt = fun.fopt

        if funId == "undefined":
            try:
                self.funId = fun.funId
            except AttributeError:
                self.funId = funId  # TODO
        else:
            self.funId = funId

        if iinstance == "undefined":
            try:
                self.iinstance = fun.iinstance
            except AttributeError:
                self.iinstance = iinstance
        else:
            self.iinstance = iinstance

        if hasattr(self, "fun"):
            self._is_samefun = self.fun == fun

        self._fun = fun
        if hasattr(fun, "_evalfull") and callable(fun._evalfull):
            self._fun_evalfull = fun._evalfull
        elif callable(fun):
            self._fun_evalfull = fun
        self.precision = dftarget
        self.fopt = fopt
        ftarget = fopt + dftarget
        self.datafileprefix = os.path.join(
            self.datapath, "data_f%s" % str(self.funId), self.fileprefix
        )
        self.indexfile = os.path.join(
            self.datapath, "%s_f%s.info" % (self.fileprefix, str(self.funId))
        )
        self._is_setfun = True
        self._fun_kwargs = kwargs
        # TODO: deal with *args and **kwargs... print everything out!

        # return self.evalfun, ftarget
        return self

    fun = property(_getfun, setfun)

    def _setdim(self, dim):
        """Sets the dimension.

        _setdim and setfun are fundamentally different:
        setfun is called by the user, whereas _setdim is called at the
        first call to method evalfun.

        """
        if hasattr(self, "_dim"):
            self._is_samedim = self._dim == dim
            if not self._is_samedim and not self._is_finalized:
                self.finalizerun()
        self._is_setdim = True
        self._dim = dim
        self.lasteval = self.__Eval()
        self.idxEvalsTrigger = 0.0
        self.evalsTrigger = 1.0
        self.nbFirstEvalsToAlwaysWrite = eval(
            nb_evaluations_always_written.replace("dim", str(dim))
        )  # TODO:
        self.idxDIMEvalsTrigger = 0.0
        self.fTrigger = np.inf
        self.idxFTrigger = np.inf

    def _getfileprefix(self):
        """String prefix common to all files generated."""
        return self._fileprefix

    def _setfileprefix(self, fileprefix):
        self._fileprefix = str(fileprefix)  # needs to be a string

    def restart(self, restart_reason="restarted"):
        """Adds an output line to the restart-log. Call this if restarts occur within run_(your)_optimizer."""
        if self._getevaluations > 0:
            buffr = []
            buffr.append(self.lasteval.sprintData(self.fopt))
            buffr.append("% restart: " + restart_reason + "\n")
            if buffr:
                fr = open(self.rdatafile, "a")
                fr.writelines(buffr)
                fr.close()

    fileprefix = property(_getfileprefix, _setfileprefix)


if __name__ == "__main__":
    print("  only one doctest implemented")
    # NotImplementedError('no doctests implemented')
    import doctest

    doctest.testmod()  # run all doctests in this module
    print("  done")



import sys

import numpy as np

sys.path.insert(0, "../")

from bayes_optim import BO, DiscreteSpace, IntegerSpace, RealSpace
from bayes_optim.surrogate import GaussianProcess, RandomForest

np.random.seed(42)


def obj_fun(x):
    x_r, x_i, x_d = np.array(x[:2]), x[2], x[3]
    if x_d == "OK":
        tmp = 0
    else:
        tmp = 1
    return np.sum((x_r + np.array([2, 2])) ** 2) + abs(x_i - 10) * 10 + tmp


def test_warm_data_with_GPR():
    dim = 2
    lb, ub = -5, 5

    def fitness(x):
        x = np.asarray(x)
        return np.sum(x ** 2)

    X = np.random.rand(5, dim) * (ub - lb) + lb
    y = [fitness(x) for x in X]
    space = RealSpace([lb, ub]) * dim

    thetaL = 1e-10 * (ub - lb) * np.ones(dim)
    thetaU = 10 * (ub - lb) * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )
    opt = BO(
        search_space=space,
        obj_fun=fitness,
        model=model,
        warm_data=(X, y),
        max_FEs=10,
        verbose=True,
        n_point=1,
    )
    assert np.all(np.asarray(opt.data) == np.asarray(opt.warm_data))
    assert opt.model.is_fitted
    opt.run()


def test_warm_data_with_RF():
    space = (
        RealSpace([-10, 10]) * 2
        + IntegerSpace([5, 15])
        + DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"])
    )

    X = space.sample(10)
    y = [obj_fun(x) for x in X]

    model = RandomForest(levels=space.levels)
    opt = BO(
        search_space=space,
        obj_fun=obj_fun,
        model=model,
        minimize=True,
        eval_type="list",
        max_FEs=5,
        verbose=True,
        acquisition_fun="EI",
        warm_data=(X, y),
    )
    opt.run()
    assert opt.data.shape[0] == 15



import logging
import time
from typing import Callable, List, Tuple

import numpy as np
from scipy.optimize import fmin_l_bfgs_b

from ..search_space import RealSpace
from ..utils import dynamic_penalty
from .mies import MIES
from .one_plus_one_cma_es import OnePlusOne_Cholesky_CMA, OnePlusOne_CMA

__all__ = ["argmax_restart", "OnePlusOne_CMA", "OnePlusOne_Cholesky_CMA", "MIES"]


def finite_difference(f: Callable, x: np.ndarray, delta: float = 1e-5, **kwargs):
    N = len(x)
    v = np.eye(N) * delta
    return np.array(
        [(f(x + v[i], **kwargs) - f(x - v[i], **kwargs)) / (2 * delta) for i in range(N)]
    ).reshape(-1, 1)


class Penalized:
    """Construct a penalized problem of form, `f(x) + p(x)`, where `p` is
    the `dynamic_penalty` function from ..utils module. We also take care of approximating
    the gradient, which is then used by the BFGS algorithm.
    """

    def __init__(self, func: Callable, h: Callable, g: Callable):
        self.func = func
        self.h = h
        self.g = g
        self.t = 10

    def __call__(self, x: np.ndarray) -> Tuple[float, np.ndarray]:
        f, fg = tuple(map(lambda x: -1.0 * x, self.func(x)))
        if self.h or self.g:
            p = dynamic_penalty(x, t=self.t, equality=self.h, inequality=self.g)
            pg = finite_difference(
                dynamic_penalty, x, t=self.t, equality=self.h, inequality=self.g
            )
            self.t += 1
            return f + p, fg + pg
        return f, fg


def argmax_restart(
    obj_func: Callable,
    search_space,
    h: Callable = None,
    g: Callable = None,
    eval_budget: int = 100,
    n_restart: int = 10,
    wait_iter: int = 3,
    optimizer: str = "BFGS",
    logger: logging.Logger = None,
):
    # lists of the best solutions and acquisition values from each restart
    xopt, fopt = [], []
    best = -np.inf
    wait_count = 0
    if not isinstance(search_space, RealSpace) and optimizer == "BFGS":
        optimizer = "MIES"
        logger.warning("L-BFGS-B can only be applied on continuous search space")

    for iteration in range(n_restart):
        x0 = search_space.sample(N=1, method="uniform")[0]
        if optimizer == "BFGS":
            bounds = np.array(search_space.bounds)
            # TODO: this is still subject to testing
            xopt_, fopt_, stop_dict = fmin_l_bfgs_b(
                Penalized(obj_func, h, g),
                x0,
                pgtol=1e-8,
                factr=1e6,
                bounds=bounds,
                maxfun=eval_budget,
            )
            xopt_ = xopt_.flatten().tolist()
            if not isinstance(fopt_, float):
                fopt_ = float(fopt_)
            fopt_ = -fopt_

            if logger is not None and stop_dict["warnflag"] != 0:
                logger.debug("L-BFGS-B terminated abnormally with the state: %s" % stop_dict)

        elif optimizer == "OnePlusOne_Cholesky_CMA":
            opt = OnePlusOne_Cholesky_CMA(
                search_space=search_space,
                obj_fun=obj_func,
                h=h,
                g=g,
                max_FEs=eval_budget,
                ftol=1e-4,
                xtol=1e-4,
                n_restart=0,
                minimize=False,
                verbose=False,
                seed=time.time(),
            )
            xopt_, fopt_, stop_dict = opt.run()
            stop_dict["funcalls"] = stop_dict["FEs"] if "FEs" in stop_dict else opt.eval_count

        elif optimizer == "MIES":
            xopt_, fopt_, stop_dict = MIES(
                search_space,
                obj_func,
                eq_func=h,
                ineq_func=g,
                max_eval=eval_budget,
                minimize=False,
                verbose=False,
                eval_type="list",
            ).optimize()

        cond_h = all(np.isclose(np.abs(h(xopt_)), 0, atol=1e-1)) if h else True
        cond_g = all(g(xopt_) <= 0) if g else True
        if cond_h and cond_g:
            if isinstance(fopt_, np.ndarray):
                fopt_ = fopt_[0]

            if fopt_ > best:
                best = fopt_
                wait_count = 0

                if logger is not None:
                    logger.debug(
                        "restart : %d - funcalls : %d - Fopt : %f"
                        % (iteration + 1, stop_dict["funcalls"], fopt_)
                    )
            else:
                wait_count += 1

            eval_budget -= stop_dict["funcalls"]
            xopt.append(xopt_)
            fopt.append(fopt_)

        if eval_budget <= 0 or wait_count >= wait_iter:
            break

    if len(xopt) == 0:
        return [], []

    idx = np.argsort(fopt)[::-1]
    return xopt[idx[0]], fopt[idx[0]]



import logging
import sys

import numpy as np

sys.path.insert(0, "../")
sys.path.insert(0, "../../pycma")

from copy import copy
from typing import Any, Callable, List, Tuple, Union

from bayes_optim import OptimizerPipeline, ParallelBO, RealSpace
from bayes_optim.surrogate import GaussianProcess, trend
from cma import CMAEvolutionStrategy, CMAOptions
from deap import benchmarks

np.random.seed(42)
Vector = List[float]
Matrix = List[Vector]


class _BO(ParallelBO):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._hist_EI = np.zeros(3)

    def ask(self, n_point=None):
        X = super().ask(n_point=n_point)
        if self.model.is_fitted:
            _criter = self._create_acquisition(fun="EI", par={}, return_dx=False)
            self._hist_EI[(self.iter_count - 1) % 3] = np.mean([_criter(x) for x in X])
        return X

    def check_stop(self):
        _delta = self._fBest_DoE - self.fopt
        if (
            self.iter_count > 1
            and np.mean(self._hist_EI[0 : min(3, self.iter_count - 1)]) < 0.01 * _delta
        ):
            self.stop_dict["low-EI"] = np.mean(self._hist_EI)

        if self.eval_count >= (self.max_FEs / 2):
            self.stop_dict["max_FEs"] = self.eval_count

        return super().check_stop()


class _CMA(CMAEvolutionStrategy):
    def __init__(
        self,
        dim: int,
        popsize: int,
        lb: Union[float, str, Vector, np.ndarray] = -np.inf,
        ub: Union[float, str, Vector, np.ndarray] = np.inf,
        ftarget: Union[int, float] = -np.inf,
        max_FEs: Union[int, str] = np.inf,
        verbose: bool = False,
        logger=None,
    ):

        inopts = {"bounds": [lb, ub], "ftarget": ftarget, "popsize": popsize}
        sigma0 = (ub - lb) / 5
        ub = np.array([ub] * dim)
        lb = np.array([lb] * dim)
        x0 = (ub - lb) * np.random.rand(dim) + lb

        super().__init__(x0=x0, sigma0=sigma0, inopts=inopts)
        self.dim = dim
        self.logger = logger
        self.max_FEs = max_FEs
        self.ftarget = ftarget
        self.verbose = verbose
        self.stop_dict = {}

    @property
    def eval_count(self):
        return self.countevals

    @property
    def iter_count(self):
        return self.countiter

    @property
    def x(self):
        return self.mean

    @x.setter
    def x(self, x):
        self.mean = copy(x)

    @property
    def Cov(self):
        return self.C

    @Cov.setter
    def Cov(self, C):
        try:
            w, B = np.linalg.eigh(C)
            if np.all(np.isreal(w)):
                self.B = B
                self.D = w ** 0.5
                self.dC = np.diag(C)
                self.C = C
        except np.linalg.LinAlgError:
            pass

    @property
    def logger(self):
        return self.logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, logging.Logger):
            self.logger = logger
            self.logger.propagate = False
            return

    def ask(self, n_point=None):
        return super().ask(number=n_point)

    def tell(self, X, y):
        super().tell(X, y)
        x, f, _ = self.best.get()
        self.logger.info("iteration {}, fopt: {}, xopt: {}".format(self.countiter, f, x))

    def check_stop(self):
        _, f, __ = self.best.get()
        if f <= self.ftarget:
            self.stop_dict["ftarget"] = f

        if self.countevals >= self.max_FEs:
            self.stop_dict["FEs"] = self.countevals

        return bool(self.stop_dict)


dim = 2
n_point = 8
max_FEs = 30 * n_point
obj_fun = lambda x: benchmarks.himmelblau(x)[0]
lb, ub = -6, 6

search_space = RealSpace([lb, ub]) * dim
mean = trend.constant_trend(dim, beta=0)  # Ordinary Kriging

# autocorrelation parameters of GPR
thetaL = 1e-10 * (ub - lb) * np.ones(dim)
thetaU = 10 * (ub - lb) * np.ones(dim)
theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

model = GaussianProcess(
    mean=mean,
    corr="squared_exponential",
    theta0=theta0,
    thetaL=thetaL,
    thetaU=thetaU,
    nugget=1e-5,
    noise_estim=False,
    optimizer="BFGS",
    wait_iter=5,
    random_start=5 * dim,
    eval_budget=100 * dim,
)

bo = _BO(
    search_space=search_space,
    obj_fun=obj_fun,
    model=model,
    eval_type="list",
    DoE_size=n_point,
    n_point=n_point,
    acquisition_fun="MGFI",
    acquisition_par={"t": 2},
    verbose=True,
    minimize=True,
)
cma = _CMA(dim=dim, popsize=n_point, lb=lb, ub=ub)


def post_BO(BO):
    xopt = np.array(BO.xopt)
    dim = BO.dim

    H = BO.model.Hessian(xopt)
    g = BO.model.gradient(xopt)[0]

    w, B = np.linalg.eigh(H)
    M = np.diag(1 / np.sqrt(w)).dot(B.T)
    H_inv = B.dot(np.diag(1 / w)).dot(B.T)
    sigma0 = np.linalg.norm(M.dot(g)) / np.sqrt(dim - 0.5)
    kwargs = {
        "x": xopt,
        "sigma": sigma0,
        "Cov": H_inv,
    }
    return kwargs


pipe = OptimizerPipeline(
    obj_fun=obj_fun, minimize=True, n_point=n_point, max_FEs=max_FEs, verbose=True
)
pipe.add(bo, transfer=post_BO)
pipe.add(cma)
pipe.run()



import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim import RealSpace
from bayes_optim.mobo import MOBO
from bayes_optim.surrogate import RandomForest

np.random.seed(123)
dim = 2
lb, ub = -10, 10


def f1(x):
    x = np.asarray(x)
    return np.sum((x + 5) ** 2)


def f2(x):
    x = np.asarray(x)
    return np.sum((x - 5) ** 2)


space = RealSpace([lb, ub]) * dim
# Bayesian optimization also uses a Surrogate model
# For mixed variable type, the random forest is typically used
model = RandomForest(levels=space.levels)

opt = MOBO(
    search_space=space,
    obj_fun=(f1, f2),
    model=model,
    DoE_size=5,
    max_FEs=100,
    verbose=True,
    acquisition_optimization={"optimizer": "OnePlusOne_Cholesky_CMA"},
)
print(opt.run())



import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim import BO, DiscreteSpace, IntegerSpace, RealSpace
from bayes_optim.surrogate import RandomForest

seed = 666
np.random.seed(seed)
dim_r = 2  # dimension of the real values


def obj_fun0(x):
    x_r, x_i, x_d = np.array(x[:dim_r]), x[2], x[3]
    _ = 0 if x_d == "OK" else 1
    return np.sum(x_r ** 2) + abs(x_i - 10) / 123.0 + _ * 2


def obj_fun(x):
    x_r = np.array([x["continuous%d" % i] for i in range(dim_r)])
    x_i = x["ordinal"]
    x_d = x["nominal"]
    _ = 0 if x_d == "OK" else 1
    return np.sum(x_r ** 2) + abs(x_i - 10) / 123.0 + _ * 2


# Continuous variables can be specified as follows:
# a 2-D variable in [-5, 5]^2
# for 2 variables, the naming scheme is continuous0, continuous1
C = RealSpace([-5, 5], var_name="continuous") * dim_r

# Equivalently, you can also use
# C = RealSpace([[-5, 5]]] * dim)
# The general usage is:
# RealSpace([[lb_1, ub_1], [lb_2, ub_2], ..., [lb_n, ub_n]])

# Integer (ordinal) variables can be specified as follows:
# The domain of integer variables can be given as with continuous ones
# var_name is optional
I = IntegerSpace([5, 15], var_name="ordinal")

# Discrete (nominal) variables can be specified as follows:
# No lb, ub... a list of categories instead
N = DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"], var_name="nominal")

# The whole search space can be constructed:
search_space = C + I + N

# Bayesian optimization also uses a Surrogate model
# For mixed variable type, the random forest is typically used
model = RandomForest(levels=search_space.levels)

opt = BO(
    search_space=search_space,
    obj_fun=obj_fun,
    model=model,
    max_FEs=50,
    DoE_size=3,  # the initial DoE size
    eval_type="dict",
    acquisition_fun="MGFI",
    acquisition_par={"t": 2},
    n_job=1,  # number of processes
    n_point=1,  # number of the candidate solution proposed in each iteration
    verbose=True,  # turn this off, if you prefer no output
)
xopt, fopt, stop_dict = opt.run()

print("xopt: {}".format(xopt))
print("fopt: {}".format(fopt))
print("stop criteria: {}".format(stop_dict))



import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim import DiscreteSpace, IntegerSpace, RealSpace
from bayes_optim.extension import MultiAcquisitionBO
from bayes_optim.surrogate import RandomForest

dim_r = 2  # dimension of the real values


def obj_fun(x):
    x_r = np.array([x["continuous%d" % i] for i in range(dim_r)])
    x_i = x["ordinal"]
    x_d = x["nominal"]
    _ = 0 if x_d == "OK" else 1
    return np.sum(x_r ** 2) + abs(x_i - 10) / 123.0 + _ * 2


search_space = (
    RealSpace([-5, 5], var_name="continuous") * dim_r
    + IntegerSpace([5, 15], var_name="ordinal")
    + DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"], var_name="nominal")
)
model = RandomForest(levels=search_space.levels)

opt = MultiAcquisitionBO(
    search_space=search_space,
    obj_fun=obj_fun,
    model=model,
    max_FEs=40,
    DoE_size=4,  # the initial DoE size
    eval_type="dict",
    n_job=4,  # number of processes
    n_point=4,  # number of the candidate solution proposed in each iteration
    verbose=True,  # turn this off, if you prefer no output
)

xopt, fopt, stop_dict = opt.run()
print("xopt: {}".format(xopt))
print("fopt: {}".format(fopt))
print("stop criteria: {}".format(stop_dict))



import os
import sys
from time import time

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

import bbobbenchmarks as bn
import fgeneric

np.random.seed(42)


class _GaussianProcessRegressor(GaussianProcessRegressor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.is_fitted = False

    def fit(self, X, y):
        super().fit(X, y)
        self.is_fitted = True
        return self

    def predict(self, X, eval_MSE=False):
        _ = super().predict(X=X, return_std=eval_MSE)

        if eval_MSE:
            y_, sd = _
            sd2 = sd ** 2
            return y_, sd2
        else:
            return _


def run_optimizer(optimizer, dim, fID, instance, logfile, lb, ub, max_FEs, data_path, bbob_opt):
    """Parallel BBOB/COCO experiment wrapper"""
    # Set different seed for different processes
    start = time()
    seed = np.mod(int(start) + os.getpid(), 1000)
    np.random.seed(seed)

    data_path = os.path.join(data_path, str(instance))
    max_FEs = eval(max_FEs)

    f = fgeneric.LoggingFunction(data_path, **bbob_opt)
    f.setfun(*bn.instantiate(fID, iinstance=instance))

    opt = optimizer(dim, f.evalfun, f.ftarget, max_FEs, lb, ub, logfile)
    opt.run()

    f.finalizerun()
    with open(logfile, "a") as fout:
        fout.write(
            "{} on f{} in {}D, instance {}: FEs={}, fbest-ftarget={:.4e}, "
            "elapsed time [m]: {:.3f}\n".format(
                optimizer,
                fID,
                dim,
                instance,
                f.evaluations,
                f.fbest - f.ftarget,
                (time() - start) / 60.0,
            )
        )


def test_BO(dim, obj_fun, ftarget, max_FEs, lb, ub, logfile):
    sys.path.insert(0, "../")
    sys.path.insert(0, "../../GaussianProcess")
    from BayesOpt import BO, DiscreteSpace, IntegerSpace, RandomForest, RealSpace
    from GaussianProcess import GaussianProcess
    from GaussianProcess.trend import constant_trend

    space = RealSpace([lb, ub]) * dim

    # kernel = 1.0 * Matern(length_scale=(1, 1), length_scale_bounds=(1e-10, 1e2))
    # model = _GaussianProcessRegressor(kernel=kernel, alpha=0, n_restarts_optimizer=30, normalize_y=False)

    mean = constant_trend(dim, beta=0)  # equivalent to Simple Kriging
    thetaL = 1e-5 * (ub - lb) * np.ones(dim)
    thetaU = 10 * (ub - lb) * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        mean=mean,
        corr="matern",
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        noise_estim=False,
        nugget=0,
        optimizer="BFGS",
        wait_iter=5,
        random_start=10 * dim,
        eval_budget=200 * dim,
    )

    return BO(
        search_space=space,
        obj_fun=obj_fun,
        model=model,
        DoE_size=dim * 10,
        max_FEs=max_FEs,
        verbose=True,
        n_point=1,
        minimize=True,
        acquisition_fun="EI",
        ftarget=ftarget,
        logger=None,
    )


if __name__ == "__main__":
    dims = (2,)
    fIDs = bn.nfreeIDs[6:]  # for all fcts
    instance = [1] * 10

    algorithm = test_BO

    opts = {"max_FEs": "50", "lb": -5, "ub": 5, "data_path": "./bbob_data/%s" % algorithm.__name__}
    opts["bbob_opt"] = {
        "comments": "max_FEs={0}".format(opts["max_FEs"]),
        "algid": algorithm.__name__,
    }

    for dim in dims:
        for fID in fIDs:
            for i in instance:
                run_optimizer(algorithm, dim, fID, i, logfile="./log", **opts)



from __future__ import annotations

from abc import ABC
from typing import Callable, Tuple

import numpy as np
import torch
from gpytorch.distributions import MultivariateNormal
# from scipy.optimize import root_scalar
from scipy.stats import norm
from torch import Tensor
from torch.nn import Module

from .._exception import ConstraintEvaluationError

__author__ = "Hao Wang"


class SCMC:
    r"""Sequential Constrained Monte Carlo sampling

    References

    .. [GolchiL15]
        Golchi, Shirin, and Jason L. Loeppky. "Monte Carlo based designs for constrained domains."
        arXiv preprint arXiv:1512.07328 (2015).

    TODO: consider adopting MIES-like self-adaptation for the MH algorithm
    """

    def __init__(
        self,
        sample_space,
        constraints: Callable,
        target_dist: Callable = lambda x: 1,
        metropolis_hastings_step: int = 15,
        tol: float = 1e-2,
    ):
        """Sequential Constrained Monte Carlo sampling

        Parameters
        ----------
        sample_space : SearchSpace
            the sample space
        constraints : Callable
            constraint functions, which could represent inequalities or equalities
        target_dist : Callable, optional
            the target probability density, by default uniform (lambda x: 1)
        metropolis_hastings_step : int, optional
            the number of MCMC iterations, by default 15
        tol : float, optional
            the tolerance on the constraint
        """
        if hasattr(constraints, "__call__"):
            constraints = [constraints]
        assert hasattr(target_dist, "__call__")

        self.constraints = constraints
        self.target_dist = target_dist
        self.sample_space = sample_space
        self.dim = sample_space.dim
        self.mh_steps = metropolis_hastings_step
        self.nu_target = tol / 8  # 8-sigma CI leads to ~1.24e-15 significance
        self.nu0 = 10
        self.nu_schedule = np.logspace(
            np.log10(self.nu0), np.log10(self.nu_target), base=10, num=15
        )

        # index of each type of variables in the dataframe
        self.id_r = self.sample_space.real_id  # index of continuous variable
        self.id_i = self.sample_space.integer_id  # index of integer variable
        self.id_d = self.sample_space.categorical_id  # index of categorical variable

        # the number of variables per each type
        self.N_r = len(self.id_r)
        self.N_i = len(self.id_i)
        self.N_d = len(self.id_d)

        self.bounds_r = np.asarray([self.sample_space.bounds[_] for _ in self.id_r])
        self.bounds_i = np.asarray([self.sample_space.bounds[_] for _ in self.id_i])
        self.bounds_d = [self.sample_space.bounds[_] for _ in self.id_d]

    def _log_posterior(self, x: np.ndarray, nu: float) -> float:
        r"""Log-posterior \log P(x) + \sum_{i=1}^K \log \Phi(-C_i(x) / \nu)

        Parameters
        ----------
        x : np.ndarray
            a point in the search space
        nu : float

        Returns
        -------
        float
            the log-probability
        """
        try:
            c = np.array([cons(x) for cons in self.constraints])
        except Exception as e:
            raise ConstraintEvaluationError(x, str(e)) from None
        return np.log(self.target_dist(x)) + np.sum(norm.logcdf(-1 * c / nu))

    def _rproposal(self, X: np.ndarray, t: int) -> np.ndarray:
        X_ = X.copy()
        self._rproposal_real(X_, t)
        self._rproposal_integer(X_, t)
        self._rproposal_discrete(X_, t)
        return X_

    def _rproposal_real(self, X: np.ndarray, t: int):
        if self.N_r > 0:
            X_ = X[:, self.id_r].astype(float)
            q = np.std(X_, axis=0) / (t + 1)
            X[:, self.id_r] = np.clip(
                X_ + q * np.random.randn(*X_.shape), self.bounds_r[:, 0], self.bounds_r[:, 1]
            )

    def _rproposal_integer(self, X: np.ndarray, t: int):
        if self.N_i > 0:
            X_ = X[:, self.id_i].astype(int)
            p = norm.cdf(np.std(X_, axis=0) / (t + 1))
            X[:, self.id_i] = np.clip(
                X_ + np.random.geometric(p, X_.shape) - np.random.geometric(p, X_.shape),
                self.bounds_i[:, 0],
                self.bounds_i[:, 1],
            )

    def _rproposal_discrete(self, X: np.ndarray, t: int):
        if self.N_d > 0:
            N = len(X)
            prob = max(0.5 * np.exp(-1 * t), 1 / self.N_d)
            for i in self.id_d:
                levels = np.array(self.sample_space.bounds[i])
                idx = np.random.rand(N) < prob
                X[idx, i] = levels[np.random.randint(0, len(levels), sum(idx))]

    def _ess(self, nu: float, nu_ref: float, X: np.ndarray) -> float:
        """effective sample size"""
        w = self.get_weights(nu, nu_ref, X)
        v = 0 if np.any(np.isnan(w)) else 1 / np.sum(w ** 2)
        return v - len(X) / 2

    def get_weights(self, nu, nu_ref, X):
        w = np.exp([self._log_posterior(x, nu) - self._log_posterior(x, nu_ref) for x in X])
        return w / np.sum(w)

    def _metropolis_hastings(self, X: np.ndarray, nu: float, t: int) -> np.ndarray:
        r"""Metropolis-Hastings algorithms to sample from a distribution given
        by `self._log_posterior`.

        Parameters
        ----------
        X : np.ndarray
            the initial sample
        nu : float
            parameter in `self._log_posterior`
        t : int
            iteration counter

        Returns
        -------
        np.ndarray
            (dependent) sample points from `self._log_posterior`
        """
        N = len(X)  # sample size
        lp = np.array([self._log_posterior(x, nu) for x in X])  # log-probability
        for _ in range(self.mh_steps):
            X_ = self._rproposal(X, t)
            for i in range(self.dim):
                X_dim = X.copy()
                X_dim[:, i] = X_[:, i]
                lp_ = np.array([self._log_posterior(x_, nu) for x_ in X_dim])
                prob = np.clip(np.exp(lp_ - lp), 0, 1)
                mask = np.random.rand(N) <= prob
                X[mask, i] = X_[mask, i]
                lp[mask] = lp_[mask]
        return X

    def sample(self, N: int) -> np.ndarray:
        """Draw a sample under constraints

        Parameters
        ----------
        N : int
            sample size

        Returns
        -------
        np.ndarray
            the sample
        """
        # draw the initial samples u.a.r. from `self.search_space`
        X = getattr(self.sample_space, "_sample")(N, method="LHS")
        for t, nu in enumerate(self.nu_schedule):
            nu_ = nu  # `nu_` -> the old `nu`
            # nu = self.nu_schedule[t]
            # if self._ess(self.nu_target, nu, X) > 0:
            #     nu *= 0.3
            # else:
            #     try:
            #         nu = (
            #             10
            #             ** root_scalar(
            #                 lambda log_nu, *args: self._ess(10 ** log_nu, *args),
            #                 args=(nu_, X),
            #                 bracket=(
            #                     np.log10(0.001 if t < 2 else self.nu_target),
            #                     np.log10(nu_),
            #                 ),
            #             ).root
            #         )
            #     except ValueError:
            #         return X
            w = self.get_weights(nu, nu_, X)
            if any(np.isnan(w)):
                w = np.ones(N) / N

            # resampling with replacement to increase the fraction of samples that are
            # more likely distributed from the new probability law, which is controlled by `nu`
            if N >= 10:
                idx = np.random.choice(N, N, p=w)
                X = X[idx, :]

            X = self._metropolis_hastings(X, nu, t)
        return X


class MCSampler(Module, ABC):
    r"""Abstract base class for Samplers.

    Attributes:
        resample: If `True`, re-draw samples in each `forward` evaluation -
            this results in stochastic acquisition functions (and thus should
            not be used with deterministic optimization algorithms).
        collapse_batch_dims: If True, collapse the t-batch dimensions of the
            produced samples to size 1. This is useful for preventing sampling
            variance across t-batches.
    """

    def __init__(self, num_samples: int, batch_range: Tuple[int, int] = (0, -2)) -> None:
        r"""Abstract base class for Samplers.

        Args:
            batch_range: The range of t-batch dimensions in the `base_sample_shape`
                used by `collapse_batch_dims`. The t-batch dims are
                batch_range[0]:batch_range[1]. By default, this is (0, -2),
                for the case where the non-batch dimensions are -2 (q) and
                -1 (d) and all dims in the front are t-batch dims.
        """
        super().__init__()
        self.batch_range = batch_range
        self.register_buffer("base_samples", None)
        self._sample_shape = torch.Size([num_samples])

    @property
    def batch_range(self) -> Tuple[int, int]:
        r"""The t-batch range."""
        return tuple(self._batch_range.tolist())

    @batch_range.setter
    def batch_range(self, batch_range: Tuple[int, int]):
        r"""Set the t-batch range and clear base samples.

        Args:
            batch_range: The range of t-batch dimensions in the `base_sample_shape`
                used by `collapse_batch_dims`. The t-batch dims are
                batch_range[0]:batch_range[1]. By default, this is (0, -2),
                for the case where the non-batch dimensions are -2 (q) and
                -1 (d) and all dims in the front are t-batch dims.
        """
        # set t-batch range if different; trigger resample & set base_samples to None
        if not hasattr(self, "_batch_range") or self.batch_range != batch_range:
            self.register_buffer("_batch_range", torch.tensor(batch_range, dtype=torch.long))
            self.register_buffer("base_samples", None)

    def forward(self, mean, variance) -> Tensor:
        r"""Draws MC samples from the posterior."""
        mvn = MultivariateNormal(
            torch.Tensor(np.array(mean)).unsqueeze(0),
            torch.Tensor([np.diag(variance[i, :]) for i in range(variance.shape[0])]).unsqueeze(0),
        )
        return mvn.rsample(sample_shape=self._sample_shape)

    @property
    def sample_shape(self) -> torch.Size:
        r"""The shape of a single sample."""
        return self._sample_shape



import sys

import numpy as np

sys.path.insert(0, "../")

from bayes_optim.surrogate import GaussianProcess, RandomForest, SurrogateAggregation
from sklearn.metrics import r2_score

np.random.seed(12)


def test_pickling():
    n_sample = 110
    levels = ["OK", "A", "B", "C", "D", "E"]
    X = np.c_[
        np.random.randn(n_sample, 2).astype(object), np.random.choice(levels, size=(n_sample, 1))
    ]
    y = np.sum(X[:, 0:-1] ** 2.0, axis=1) + 5 * (X[:, -1] == "OK")

    X_train, y_train = X[:100, :], y[:100]
    X_test, y_test = X[100:, :], y[100:]

    # sklearn-random forest
    rf = RandomForest(levels={2: levels}, max_features="sqrt")
    rf.fit(X_train, y_train)
    y_hat, mse = rf.predict(X_test, eval_MSE=True)

    print("sklearn random forest:")
    print("target :", y_test)
    print("predicted:", y_hat)
    print("MSE:", mse)
    print("r2:", r2_score(y_test, y_hat))
    print()

    X = np.c_[
        np.random.randn(n_sample, 2).astype(object),
        np.random.choice(levels, size=(n_sample, 1)),
    ]
    y = np.sum(X[:, 0:-1] ** 2.0, axis=1) + 5 * (X[:, -1] == "OK")

    X_train, y_train = X[:100, :], y[:100]
    X_test, y_test = X[100:, :], y[100:]

    rf_ = RandomForest(levels={2: levels}, max_features="sqrt")
    rf_.fit(X_train, y_train)

    rf_aggr = SurrogateAggregation((rf, rf_), weights=(0.1, 0.9))
    y_hat, mse = rf_aggr.predict(X_test, eval_MSE=True)

    print("sklearn random forest:")
    print("target :", y_test)
    print("predicted:", y_hat)
    print("MSE:", mse)
    print("r2:", r2_score(y_test, y_hat))


def test_multioutput_gpr():
    n_sample = 100
    dim = 10
    X = np.random.rand(n_sample, dim)
    y = np.sum(X ** 2.0, axis=1)
    thetaL, thetaU = 1e-10 * np.ones(dim), 10 * np.ones(dim)

    # noiseless mode
    model = GaussianProcess(
        theta0=np.random.rand(dim) * (thetaU - thetaL) + thetaL,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )
    model.fit(X, y)
    model.predict(X, eval_MSE=True)

    # noisy mode
    model = GaussianProcess(
        theta0=np.random.rand(dim) * (thetaU - thetaL) + thetaL,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=1e-6,
        noise_estim=False,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )
    model.fit(X, y)
    model.predict(X, eval_MSE=True)

    # noise estimation mode
    model = GaussianProcess(
        theta0=np.random.rand(dim) * (thetaU - thetaL) + thetaL,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=1e-6,
        noise_estim=True,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )
    model.fit(X, y)
    model.predict(X, eval_MSE=True)



from typing import List


class AskEmptyError(Exception):
    """Exception raised if `BO.ask` yields an empty outcome

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message=(
            "Ask yields empty solutions. This could be caused when "
            "when constraints are too restrict or the search space is already enumerated."
        ),
    ):
        self.message = message
        super().__init__(self.message)


class FlatFitnessError(Exception):
    """Exception raised if a flat fitness landscape is observed

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message=(
            "Too many flat objective values observed. The optimization process is terminated."
        ),
    ):
        self.message = message
        super().__init__(self.message)


class RecommendationUnavailableError(Exception):
    """Exception raised if the the recommendation is not available

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, details: str = ""):
        self.message = f"Optimizer's recommendation is not yet available due to {details}."
        super().__init__(self.message)


class ObjectiveEvaluationError(Exception):
    """Exception raised if errors occur when evaluating the objective function

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, solution: List, details: str = ""):
        self.message = (
            f"Evaluating {solution} on the objective function "
            f"gives the following error{details}. Please check the declaration of the objective."
        )
        super().__init__(self.message)


class ConstraintEvaluationError(Exception):
    """Exception raised if errors occur when evaluating the constraint function

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, solution: List, details: str = ""):
        self.message = (
            f"Evaluating {solution} with on the constraints "
            f"gives the following error: {details}. Please check the declaration of the constraint."
        )
        super().__init__(self.message)



import sys

import numpy as np

sys.path.insert(0, "./")

import matplotlib.pyplot as plt
from bayes_optim import DiscreteSpace, IntegerSpace, RealSpace
from bayes_optim.samplers import SCMC
from scipy.stats import norm

dim = 2
search_space = RealSpace([0, 1]) * dim
h = lambda x: np.abs(np.sum(x) - 1)
tol = 1e-2
sampler = SCMC(
    search_space,
    [
        h,
    ],
    tol=tol,
)

for i in range(20):
    X = sampler.sample(5)
    print(X)
    assert np.all([np.isclose(h(x), 0, atol=tol) for x in X])


dim = 2
search_space = (
    RealSpace([-5, 5]) * dim + DiscreteSpace(["A", "B", "C", "D"]) + IntegerSpace([1, 10])
)
sampler = SCMC(
    search_space,
    [
        lambda x: np.sum(x[:2] ** 2) - 1,
        lambda x: 0.25 - np.sum(x[:2] ** 2),
        lambda x: bool(x[2] not in ["A", "B"]),
        lambda x: x[3] - 5.1,
    ],
    # target_dist=lambda x: norm.pdf(x[0], scale=3) * norm.pdf(x[1]),
    tol=tol,
)
X = sampler.sample(100)
assert np.all([x[2] in ["A", "B"] for x in X])

plt.plot(X[:, 0], X[:, 1], "r.")
circle1 = plt.Circle((0, 0), 1, color="k", fill=False)
circle2 = plt.Circle((0, 0), 0.5, color="k", fill=False)

ax = plt.gca()
ax.add_patch(circle1)
ax.add_patch(circle2)
plt.show()



import functools
import time
from typing import Callable, Dict, List, Union

import numpy as np

from ._exception import ConstraintEvaluationError
from .solution import Solution


def is_pareto_efficient(fitness, return_mask: bool = True) -> Union[List[int], List[bool]]:
    """get the Pareto efficient subset

    Parameters
    ----------
    fitness : np.ndarray of shape (n_points, n_obj)
        the objective value
    return_mask : bool, optional
        if returning a mask, by default True

    Returns
    -------
    An array of indices of pareto-efficient points.
        If return_mask is True, this will be an (n_points, ) boolean array
        Otherwise it will be a (n_efficient_points, ) integer array of indices.
    """

    is_efficient = np.arange(fitness.shape[0])
    n_points = fitness.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(fitness):
        nondominated_point_mask = np.any(fitness < fitness[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        fitness = fitness[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1

    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype=bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    return is_efficient


def fillin_fixed_value(X: List[List], fixed: Dict, search_space):
    if fixed is None:
        return X
    if len(X) == 0:
        return X
    mask = np.array([v in fixed.keys() for v in search_space.var_name])
    values = [fixed[k] for i, k in enumerate(search_space.var_name) if mask[i]]
    out = np.empty((len(X), len(mask)), dtype=object)
    out[:, mask] = values
    out[:, ~mask] = X
    return out.tolist()


def partial_argument(
    func: callable,
    var_name: List[str],
    fixed: Dict[str, Union[str, float, int, object, bool]] = None,
    reduce_output: bool = False,
):
    """fill-in the values for inactive variables

    Parameters
    ----------
    func : callable
        the target function to call which is defined on the original search space
    masks : np.ndarray
        the mask array indicating which variables are deemed inactive
    values : np.ndarray
        the values fixed for the inactive variables
    """
    fixed = {} if fixed is None else fixed
    masks = np.array([v in fixed.keys() for v in var_name])
    values = [fixed[k] for i, k in enumerate(var_name) if masks[i]]

    @functools.wraps(func)
    def wrapper(X: Union[np.ndarray, Solution, list]):
        if not isinstance(X, np.ndarray):
            X = np.array(X, dtype=object)

        N = 1 if len(X.shape) == 1 else X.shape[1]
        X_ = np.empty((N, len(masks)), dtype=object)
        X_[:, masks] = values
        X_[:, ~masks] = X
        out_ = func(X_)

        # TODO: fix this ad-hoc solution for acquisition functions
        if reduce_output:
            out = []
            for v in tuple(out_):
                if isinstance(v, np.ndarray):
                    if len(v.shape) == 1 and len(v) > 1:
                        v = v[~masks]
                    elif len(v.shape) == 2:
                        if v.shape[0] == len(masks):
                            v = v[~masks, :]
                        elif v.shape[1] == len(masks):
                            v = v[:, ~masks]
                elif isinstance(v, list) and len(v) == len(masks):
                    v = [v[m] for m in ~masks]
                out.append(v)
            return tuple(out)
        return out_

    return wrapper


def func_with_list_arg(func, arg_type, var_names):
    @functools.wraps(func)
    def wrapper(X):
        if isinstance(X, (list, tuple)):
            X = np.array(X, dtype="object")
        if len(X.shape) == 1:
            X = X[np.newaxis, :]
        X = Solution(X, var_name=var_names)
        if arg_type == "list":
            X = X.tolist()
        elif arg_type == "dict":
            X = X.to_dict()
        return np.array([func(_) for _ in X]).ravel()

    return wrapper


def timeit(func):
    @functools.wraps(func)
    def __func__(ref, *arg, **kwargv):
        t0 = time.time()
        out = func(ref, *arg, **kwargv)
        if hasattr(ref, "logger"):
            ref.logger.info(f"{func.__name__} takes {time.time() - t0:.4f}s")
        else:
            print(f"{func.__name__} takes {time.time() - t0:.4f}s")
        return out

    return __func__


def arg_to_int(arg):
    if isinstance(arg, str):
        x = int(eval(arg))
    elif isinstance(arg, (int, float)):
        x = int(arg)
    else:
        raise ValueError
    return x


def set_bounds(bound, dim):
    if isinstance(bound, str):
        bound = eval(bound)
    elif isinstance(bound, (float, int)):
        bound = [bound] * dim
    elif hasattr(bound, "__iter__"):
        bound = list(bound)
        if len(bound) == 1:
            bound *= dim
    assert len(bound) == dim
    return np.asarray(bound)


def dynamic_penalty(
    X: List,
    t: int = 1,
    equality: Callable = None,
    inequality: Callable = None,
    C: float = 0.5,
    alpha: float = 1,
    beta: float = 1.5,
    epsilon: float = 1e-1,
    minimize: bool = True,
) -> np.ndarray:
    r"""Dynamic Penalty calculated as follows:

    $$(tC)^{\alpha} * [\sum_i max(|h(x_i)|, \epsilon) + \sum_i max(0, g(x_i))^{\beta}],$$

    where $x_i$ -> each row of ``X``, h -> ``equality``, and g -> ``inequality``.

    TODO: give a reference here

    Parameters
    ----------
    X : np.ndarray
        Input candidate solutions
    t : int, optional
        The iteration number of the optimization algorithm employing this method, by default 1
    equality : Callable, optional
        Equality function, by default None
    inequality : Callable, optional
        Inequality function, by default None
    C : float, optional
        coefficient of the iteration term, by default 0.5
    alpha : float, optional
        exponent to the iteration term, by default 1
    beta : float, optional
        coefficient to the inequality terms, by default 2
    epsilon : float, optional
        threshold to determine whether the equality constraint is met, by default 1e-4
    minimize : bool, optional
        minimize or maximize? by default True

    Returns
    -------
    ``p``
        the dynamic penalty value
    """
    if not hasattr(X[0], "__iter__") or isinstance(X[0], str):
        X = [X]
    X = np.array(X, dtype=object)

    N = len(X)
    p = np.zeros(N)

    if equality is not None:
        try:
            v = np.atleast_2d(list(map(equality, X))).reshape(N, -1)
        except Exception as e:
            raise ConstraintEvaluationError(X, str(e)) from None
        v[np.abs(v) <= epsilon] = 0
        p += np.sum(np.abs(v), axis=1)

    if inequality is not None:
        try:
            v = np.atleast_2d(list(map(inequality, X))).reshape(N, -1)
        except Exception as e:
            raise ConstraintEvaluationError(X, str(e)) from None
        # NOTE: inequalities are always tested with less or equal relation.
        # Inequalities with strict less conditions should be created by adding a tiny epsilon
        # to the constraint
        v[v <= 0] = 0
        p += np.sum(np.abs(v) ** beta, axis=1)

    p = (C * t) ** alpha * p * (-1) ** (not minimize)
    return p



import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim.extension import PCABO, RealSpace

np.random.seed(123)
dim = 5
lb, ub = -5, 5


def fitness(x):
    x = np.asarray(x)
    return np.sum((np.arange(1, dim + 1) * x) ** 2)


space = RealSpace([lb, ub]) * dim
opt = PCABO(
    search_space=space,
    obj_fun=fitness,
    DoE_size=5,
    max_FEs=40,
    verbose=True,
    n_point=1,
    n_components=0.95,
    acquisition_optimization={"optimizer": "BFGS"},
)
print(opt.run())



from .non_dominated import FastNondominatedPartitioning, NondominatedPartitioning
from .utils import compute_dominated_hypercell_bounds_2d, compute_non_dominated_hypercell_bounds_2d

__all__ = [
    "compute_dominated_hypercell_bounds_2d",
    "compute_non_dominated_hypercell_bounds_2d",
    "FastNondominatedPartitioning",
    "NondominatedPartitioning",
]



from .node import Node
from .search_space import (
    BoolSpace,
    DiscreteSpace,
    IntegerSpace,
    OrdinalSpace,
    RealSpace,
    SearchSpace,
    SubsetSpace,
)
from .variable import Bool, Discrete, Integer, Ordinal, Real, Subset, Variable

__all__ = [
    "Node",
    "BoolSpace",
    "DiscreteSpace",
    "IntegerSpace",
    "OrdinalSpace",
    "RealSpace",
    "SearchSpace",
    "Real",
    "Integer",
    "Discrete",
    "Ordinal",
    "Bool",
    "Subset",
    "SubsetSpace",
    "Variable",
]



import sys

import numpy as np

sys.path.insert(0, "./")
from bayes_optim.acquisition_optim import OnePlusOne_Cholesky_CMA


def obj_fun(x):
    return np.sum(x ** 2)


opt = OnePlusOne_Cholesky_CMA(30, obj_fun, lb=-100, ub=100, sigma0=40, ftarget=1e-8, verbose=True)

opt.run()
print(opt.stop_dict)



from __future__ import annotations

from typing import Dict, List, Optional, Sequence, Union

import numpy as np
from tabulate import tabulate

__authors__ = ["Hao Wang"]

# TODO: maybe set var_name as a constant attribute
# TODO: register the objective function the class and implement an eval function
# TODO: fix the maximum recursion problem when debugging
# TODO: make this class work with numpy's ufunc, e.g., ``np.round``


class Solution(np.ndarray):
    """Subclassing numpy array to represent set of solutions in the optimization
    Goal to achieve:
     1) heterogenous data types, like pandas
     2) easy indexing as np.ndarray
     3) extra attributes (e.g., fitness) sliced together with the solution
    """

    def __new__(
        cls,
        x: Sequence,
        fitness: Optional[Sequence] = None,
        n_eval: Union[int, List[int]] = 0,
        index: Union[str, List[str]] = None,
        var_name: Union[str, List[str]] = None,
        fitness_name: Union[str, List[str]] = None,
        n_obj: int = 1,
        verbose: bool = True,
    ):
        """
        Parameters
        ----------
        x : Sequence
            The parameters of a solution
        fitness : Optional[Sequence], optional
            objective values of solutions, by default None
        n_eval : Union[int, List[int]], optional
            number of evaluations per each solution, by default 0
        index : Union[int, List[int]], optional
            indices of solutions, by default None
        var_name : Union[str, List[str]], optional
            names of the variables, by default None
        fitness_name : Union[str, List[str]], optional
            names of the fitness values, by default None
        n_obj : int, optional
            the number of objective functions, by default 1
        verbose : bool, optional
            controls if additional information are printed when calling ``__str__``
                and to_dict, by default True

        Returns
        -------
        ``Solution``

        Note
        ----
            Instead of using `__init__`, the `__new__` function is used here because
            sometimes we would like to return an object of its subclasses, e.g., when
            slicing a subclass of `ndarray`, `ndarray.__new__(subclass, ...)` will
            return an object of type `subclass` while `ndarray.__init__(self, ...)`
            will return an object of `ndarray` (of course, `__init__` would work if
            the user also overloads the slicing function, which is not convenient).
            If attributes `index`, `fitness`, `n_eval` are modified in a slice of
            Solution, the corresponding attributes in the original object are also
            modified. Here, `var_name` is not affected by this behavior. This function
            is only called when explicitly constructing the `Solution` object.
            For slicing and view casting, the extra attributes are handled in function
            `__array_finalize__`.
        """
        obj = np.asarray(x, dtype="object").view(cls)

        if len(obj.shape) > 2:
            raise Exception("More than 2D is not supported")

        obj.N = 1 if len(obj.shape) == 1 else obj.shape[0]
        obj.dim = obj.shape[0] if len(obj.shape) == 1 else obj.shape[1]
        obj.n_obj = int(n_obj)

        # TODO: convert those as decorators
        if obj.n_obj > 1:  # multi-objective
            if not hasattr(fitness, "__iter__"):
                fitness = [[fitness] * obj.n_obj] * obj.N
            elif not hasattr(fitness[0], "__iter__"):
                assert len(fitness) == obj.n_obj
                fitness = [fitness] * obj.N
            assert all([len(_) == obj.n_obj for _ in fitness])
        elif obj.n_obj == 1:
            if not hasattr(fitness, "__iter__"):
                fitness = [fitness] * obj.N
        assert len(fitness) == obj.N

        if not hasattr(n_eval, "__iter__"):
            assert isinstance(n_eval, int)
            n_eval = [n_eval] * obj.N
        else:
            assert all(list(map(lambda n: isinstance(n, int), n_eval)))

        if index is None:
            index = list(map(str, range(obj.N)))
        elif isinstance(index, (str, int)):
            index = [index]
        assert len(index) == obj.N

        if var_name is None:
            if obj.dim == 1:
                var_name = ["x"]
            else:
                var_name = ["x" + str(i) for i in range(obj.dim)]
        assert len(var_name) == obj.dim

        if fitness_name is None:
            if obj.n_obj == 1:
                fitness_name = ["f"]
            else:
                fitness_name = ["f" + str(i) for i in range(obj.n_obj)]
        elif isinstance(fitness_name, str):
            assert obj.n_obj == 1
            fitness_name = [fitness_name]
        assert len(fitness_name) == obj.n_obj

        # TODO: a np.ndarray is used for those attributes because slicing it returns references
        super(Solution, obj).__setattr__("fitness", np.asarray(fitness, dtype=float))
        super(Solution, obj).__setattr__("n_eval", np.asarray(n_eval, dtype=int))
        super(Solution, obj).__setattr__("index", np.asarray(index, dtype=str))
        obj.var_name = np.asarray(var_name)
        obj.fitness_name = fitness_name
        obj.verbose = verbose
        return obj

    def __iadd__(self, other: Solution) -> Solution:
        return self.__add__(other)

    def __add__(self, other: Solution) -> Solution:
        """Concatenate two Solution objects"""
        assert isinstance(other, Solution)
        assert self.dim == other.dim
        assert self.n_obj == other.n_obj
        assert len(set(self.fitness_name).symmetric_difference(other.fitness_name)) == 0
        assert len(set(self.var_name).symmetric_difference(other.var_name)) == 0

        _ = [self.tolist()] if len(self.shape) == 1 else self.tolist()
        __ = [other.tolist()] if len(other.shape) == 1 else other.tolist()
        return Solution(
            _ + __,
            self.fitness.tolist() + other.fitness.tolist(),
            self.n_eval.tolist() + other.n_eval.tolist(),
            var_name=self.var_name,
            fitness_name=self.fitness_name,
            index=self.index.tolist() + other.index.tolist(),
            n_obj=self.n_obj,
            verbose=self.verbose,
        )

    def __mul__(self, N: int) -> Solution:
        """repeat a solution N times"""
        assert isinstance(N, int)
        if self.N > 1:
            raise Exception("Replication is not supported for 2D")
        return Solution(
            [self.tolist()] * N,
            self.fitness.tolist() * N,
            self.n_eval.tolist() * N,
            var_name=self.var_name,
            fitness_name=self.fitness_name,
            n_obj=self.n_obj,
            verbose=self.verbose,
        )

    def __rmul__(self, N: int) -> Solution:
        return self.__mul__(N)

    def __setattr__(self, name, value):
        attr = getattr(self, name, None)
        if hasattr(attr, "__iter__") and name in ["fitness", "n_eval", "index"]:
            attr[:] = value  # NOTE: copy the value (not reference) to the attribute
        else:
            super(Solution, self).__setattr__(name, value)

    def __str__(self) -> str:
        var_name = self.var_name.tolist()
        headers = var_name + ["n_eval"] + self.fitness_name if self.verbose else var_name
        if len(self.shape) == 1:
            t = (
                [self.tolist() + self.n_eval.tolist() + self.fitness.tolist()]
                if self.verbose
                else [self.tolist()]
            )
        else:
            t = np.c_[self, self.n_eval, self.fitness].tolist() if self.verbose else self.tolist()

        return tabulate(t, headers=headers, showindex=self.index.tolist(), tablefmt="grid")

    def __repr__(self) -> str:
        return self.__str__()

    def __getitem__(self, index):
        _, __ = index, slice(None, None)
        if isinstance(index, tuple):
            _ = index[0]
            if len(index) == 2:
                if isinstance(index[1], int) and not isinstance(index[0], int):
                    __ = slice(index[1], index[1] + 1)
                    index = (_, __)
                else:
                    __ = index[1]

        subarr = super().__getitem__(index)

        # sub-slicing the attributes
        if isinstance(subarr, Solution):
            # NOTE: `slice` is needed here to make sure an array is always returned
            # after slicing attributes `fitness`, `n_eval, `index`
            # Otherwise setting the attribute of a slice is by value...
            _ = slice(_, _ + 1) if isinstance(_, (int, np.int_)) else _
            if len(self.shape) == 1:  # `self` is a 1-d array
                subarr.var_name = subarr.var_name[_]
            else:
                # NOTE: 1-d array should have 1-d `fitness`
                fitness = subarr.fitness[_]
                if len(subarr.shape) == 1:
                    fitness = fitness.ravel()

                super(Solution, subarr).__setattr__("fitness", fitness)
                super(Solution, subarr).__setattr__("n_eval", subarr.n_eval[_])
                super(Solution, subarr).__setattr__("index", subarr.index[_])
                super(Solution, subarr).__setattr__("var_name", subarr.var_name[__])

            subarr.N = 1 if len(subarr.shape) == 1 else subarr.shape[0]
            subarr.dim = subarr.shape[0] if len(subarr.shape) == 1 else subarr.shape[1]

        return subarr

    def __array_finalize__(self, obj: Solution):
        """
        `__array_finalize__` is called after new `Solution` instance is created: from calling
        1) `__new__`, 2) view casting (`ndarray`.`view()`) or 3) slicing (`__getitem__`)
        """
        if obj is None:
            return
        # Needed for array slicing (__getitem__)
        super(Solution, self).__setattr__("fitness", getattr(obj, "fitness", None))
        super(Solution, self).__setattr__("n_eval", getattr(obj, "n_eval", None))
        super(Solution, self).__setattr__("index", getattr(obj, "index", None))

        # NOTE: to get rid of pylint errors..
        self.index: np.ndarray = getattr(obj, "index", None)
        self.var_name: np.ndarray = getattr(obj, "var_name", None)
        self.fitness: np.ndarray = getattr(obj, "fitness", None)
        self.fitness_name: List[str] = getattr(obj, "fitness_name", None)
        self.n_eval: np.ndarray = getattr(obj, "n_eval", None)
        self.verbose: bool = getattr(obj, "verbose", None)
        self.dim: int = getattr(obj, "dim", None)
        self.N: int = getattr(obj, "N", None)
        self.n_obj: int = getattr(obj, "n_obj", None)

    @classmethod
    def from_dict(
        cls,
        x: Union[List[dict], dict],
        index: List[int, str] = None,
        var_name: Union[str, List[str]] = None,
        **kwargs,
    ) -> Solution:
        if isinstance(x, dict):
            var_name = list(x.keys())
            res = cls.__new__(cls, x=list(x.values()), var_name=var_name, index=index, **kwargs)
        elif isinstance(x, list):
            var_name = list(x[0].keys())
            _x = [list(_.values()) for _ in x]
            res = cls.__new__(cls, x=_x, var_name=var_name, index=index, **kwargs)
        return res

    def unique(self) -> Solution:
        if len(self.shape) == 1:
            return self
        _, index = np.unique(self.tolist(), axis=0, return_index=True)
        return self[np.sort(index)]

    def to_list(self) -> List:
        return super().tolist()

    def to_dict(self, orient: str = "index", with_index: bool = False) -> Union[List, Dict]:
        """A Solution object to a dictionary

        Parameters
        ----------
        orient : str, optional
            orientation of the dictionary, which is either `index` or `var`, by default "index"
        with_index : bool, optional
            whether indices of solutions should be used in the dictionary, by default False
        """
        if orient not in ("index", "var"):
            raise ValueError(f"`orient` value {orient} is invalid")

        obj = np.atleast_2d(self.view(np.ndarray))
        if orient == "index":
            if with_index:
                res = {
                    _index: {self.var_name[k]: obj[i, k] for k in range(self.dim)}
                    for i, _index in enumerate(self.index)
                }
            else:
                res = [
                    {self.var_name[k]: obj[i, k] for k in range(self.dim)}
                    for i, _index in enumerate(self.index)
                ]
                if len(self.shape) == 1:
                    res = res[0]
        elif orient == "var":
            if with_index:
                res = {
                    _name: {index: obj[i, k] for i, index in enumerate(self.index)}
                    for k, _name in enumerate(self.var_name)
                }
            else:
                res = {_name: list(obj[:, k]) for k, _name in enumerate(self.var_name)}
        return res

    def to_csv(
        self,
        fname: str,
        delimiter: str = ",",
        append: bool = False,
        header: bool = True,
        index: bool = True,
        attribute: bool = True,
    ):
        """Convert to the CSV format

        Parameters
        ----------
        fname : str
            the name of the output CSV file
        delimiter : str, optional
            column-separating delimiter, by default ","
        append : bool, optional
            whether to append to an existing file, by default False
        header : bool, optional
            whether to write a header in the CSV, by default True
        index : bool, optional
            wheter to include the index, by default True
        attribute : bool, optional
            whether to include the atttributes, i.e., `self.n_eval` and `self.fitness`,
            by default True
        """
        var_name = self.var_name.tolist()
        if header:
            _header = var_name
            if index:
                _header = [""] + _header
            if attribute:
                attr_name = ["n_eval"] + self.fitness_name
                _header += attr_name
            _header = delimiter.join(_header) + "\n"

        data = self.reshape(1, -1) if len(self.shape) == 1 else self
        if index:
            data = np.c_[self.index, data]
        if attribute:
            data = np.c_[data, self.n_eval, self.fitness]

        out = [delimiter.join(map(str, row)) + "\n" for row in data.tolist()]
        mode = "a" if append else "w"
        with open(fname, mode) as f:
            if header:
                f.writelines(_header)
            f.writelines(out)



import os
import string
import sys

sys.path.insert(0, "../")
import numpy as np
import pytest
from bayes_optim import BO, ParallelBO
from bayes_optim._exception import AskEmptyError, FlatFitnessError
from bayes_optim.search_space import (BoolSpace, DiscreteSpace, IntegerSpace,
                                      OrdinalSpace, RealSpace)
from bayes_optim.surrogate import GaussianProcess, RandomForest, trend

np.random.seed(123)


def test_pickling():
    dim = 5
    lb, ub = -1, 5

    def fitness(x):
        x = np.asarray(x)
        return np.sum(x ** 2)

    space = RealSpace([lb, ub]) * dim

    mean = trend.constant_trend(dim, beta=None)
    thetaL = 1e-10 * (ub - lb) * np.ones(dim)
    thetaU = 10 * (ub - lb) * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        mean=mean,
        corr="squared_exponential",
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )
    opt = BO(
        search_space=space,
        obj_fun=fitness,
        model=model,
        DoE_size=5,
        max_FEs=10,
        verbose=True,
        n_point=1,
        logger="log",
    )
    opt.save("test")
    opt = BO.load("test")

    print(opt.run())

    os.remove("test")
    os.remove("log")


@pytest.mark.parametrize("var_type", ["r", "b", "c", "i", "o"])
def test_homogenous(var_type):
    dim = 5

    def fitness(_):
        return np.random.rand()

    if var_type == "r":
        lb, ub = -1, 5
        space = RealSpace([lb, ub]) * dim
        mean = trend.constant_trend(dim, beta=None)
        thetaL = 1e-10 * (ub - lb) * np.ones(dim)
        thetaU = 10 * (ub - lb) * np.ones(dim)
        theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

        model = GaussianProcess(
            mean=mean,
            corr="squared_exponential",
            theta0=theta0,
            thetaL=thetaL,
            thetaU=thetaU,
            nugget=0,
            noise_estim=False,
            optimizer="BFGS",
            wait_iter=3,
            random_start=dim,
            likelihood="concentrated",
            eval_budget=100 * dim,
        )
    else:
        if var_type == "b":
            space = BoolSpace() * dim
        elif var_type == "i":
            space = IntegerSpace([0, 10], step=1) * dim
        elif var_type == "c":
            space = DiscreteSpace(list(range(10))) * dim
        elif var_type == "o":
            space = OrdinalSpace(list(string.ascii_lowercase))
        model = RandomForest(levels=space.levels)

    opt = BO(
        search_space=space,
        obj_fun=fitness,
        model=model,
        DoE_size=5,
        max_FEs=10,
        verbose=True,
        n_point=1,
    )
    print(opt.run())


def test_fixed_var():
    def obj_fun(x):
        return (
            x["continuous"] ** 2
            + abs(x["ordinal"] - 10) / 123.0
            + (x["nominal"] != "OK") * 2
            + int(x["bool"]) * 3
        )

    search_space = (
        BoolSpace(var_name="bool")
        + IntegerSpace([5, 15], var_name="ordinal")
        + RealSpace([-5, 5], var_name="continuous")
        + DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"], var_name="nominal")
    )
    opt = ParallelBO(
        search_space=search_space,
        obj_fun=obj_fun,
        model=RandomForest(levels=search_space.levels),
        max_FEs=100,
        DoE_size=3,  # the initial DoE size
        eval_type="dict",
        acquisition_fun="MGFI",
        acquisition_par={"t": 2},
        n_job=1,  # number of processes
        n_point=3,  # number of the candidate solution proposed in each iteration
        verbose=True,  # turn this off, if you prefer no output
    )
    X = opt.ask(3, fixed={"ordinal": 5, "continuous": 3.2})
    assert all([x["ordinal"] == 5 and x["continuous"] == 3.2 for x in X])
    opt.tell(X, [obj_fun(x) for x in X])
    X = opt.ask(3, fixed={"nominal": "OK", "bool": False})
    assert all([x["nominal"] == "OK" and not x["bool"] for x in X])


def test_flat_continuous():
    dim = 5
    lb, ub = -1, 5

    def fitness(_):
        return 1

    space = RealSpace([lb, ub]) * dim

    mean = trend.constant_trend(dim, beta=None)
    thetaL = 1e-10 * (ub - lb) * np.ones(dim)
    thetaU = 10 * (ub - lb) * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        mean=mean,
        corr="squared_exponential",
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )
    opt = BO(
        search_space=space,
        obj_fun=fitness,
        model=model,
        DoE_size=5,
        max_FEs=10,
        verbose=True,
        n_point=1,
    )
    with pytest.raises(FlatFitnessError):
        opt.run()


def test_infeasible_constraints():
    dim = 5
    lb, ub = -5, 5

    def fitness(_):
        return 1

    space = RealSpace([lb, ub]) * dim
    model = RandomForest(levels=space.levels)
    opt = BO(
        search_space=space,
        obj_fun=fitness,
        model=model,
        DoE_size=5,
        ineq_fun=lambda x: x[0] + 5.1,
        max_FEs=10,
        verbose=True,
        n_point=1,
    )
    with pytest.raises(AskEmptyError):
        opt.run()


@pytest.mark.parametrize("eval_type", ["list", "dict"])  # type: ignore
def test_mix_space(eval_type):
    dim_r = 2  # dimension of the real values
    if eval_type == "dict":

        def obj_fun(x):
            # Do explicit type-casting since dataframe rows might be strings otherwise
            x_r = np.array([float(x["continuous%d" % i]) for i in range(dim_r)])
            x_i = int(x["ordinal"])
            x_d = x["nominal"]
            _ = 0 if x_d == "OK" else 1
            return np.sum(x_r ** 2) + abs(x_i - 10) / 123.0 + _ * 2

    elif eval_type == "list":

        def obj_fun(x):
            x_r = np.array([x[i] for i in range(dim_r)])
            x_i = x[-2]
            x_d = x[-1]
            _ = 0 if x_d == "OK" else 1
            return np.sum(x_r ** 2) + abs(x_i - 10) / 123.0 + _ * 2

    else:
        raise NotImplementedError

    search_space = (
        RealSpace([-5, 5], var_name="continuous") * dim_r
        + IntegerSpace([5, 15], var_name="ordinal")
        + DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"], var_name="nominal")
    )

    model = RandomForest(levels=search_space.levels)

    opt = ParallelBO(
        search_space=search_space,
        obj_fun=obj_fun,
        model=model,
        max_FEs=6,
        DoE_size=3,  # the initial DoE size
        eval_type=eval_type,
        acquisition_fun="MGFI",
        acquisition_par={"t": 2},
        n_job=1,  # number of processes
        n_point=3,  # number of the candidate solution proposed in each iteration
        verbose=True,  # turn this off, if you prefer no output
    )
    xopt, fopt, stop_dict = opt.run()

    print("xopt: {}".format(xopt))
    print("fopt: {}".format(fopt))
    print("stop criteria: {}".format(stop_dict))



import functools
import logging
import os
import sys
from abc import ABC, abstractmethod
from copy import copy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import dill
import numpy as np
from joblib import Parallel, delayed
from sklearn.metrics import mean_absolute_percentage_error, r2_score

from . import acquisition_fun as AcquisitionFunction
from ._exception import AskEmptyError, FlatFitnessError, RecommendationUnavailableError
from .acquisition_optim import argmax_restart
from .acquisition_optim.option import (
    default_AQ_max_FEs,
    default_AQ_n_restart,
    default_AQ_wait_iter,
)
from .misc import LoggerFormatter
from .search_space import RealSpace, SearchSpace
from .solution import Solution
from .utils import (
    arg_to_int,
    dynamic_penalty,
    fillin_fixed_value,
    func_with_list_arg,
    partial_argument,
    timeit,
)

__authors__ = ["Hao Wang"]


class BaseBO(ABC):
    """Bayesian Optimization base class, which implements the Ask-Evaluate-Tell interface"""

    def __init__(
        self,
        search_space: SearchSpace,
        obj_fun: Optional[Callable] = None,
        parallel_obj_fun: Optional[Callable] = None,
        eq_fun: Optional[Callable] = None,
        ineq_fun: Optional[Callable] = None,
        model: Optional[Any] = None,
        eval_type: str = "list",
        DoE_size: Optional[int] = None,
        warm_data: Tuple = (),
        n_point: int = 1,
        acquisition_fun: str = "EI",
        acquisition_par: dict = {},
        acquisition_optimization: dict = {},
        ftarget: Optional[float] = None,
        max_FEs: Optional[int] = None,
        minimize: bool = True,
        n_job: int = 1,
        data_file: Optional[str] = None,
        verbose: bool = False,
        random_seed: Optional[int] = None,
        logger: Optional[str] = None,
        instance_name: Optional[str] = None,
        **kwargs,
    ):
        """The base class for Bayesian Optimization

        Parameters
        ----------
        search_space : SearchSpace
            The search space, an instance of `SearchSpace` class.
        obj_fun : Callable
            The objective function to optimize.
        parallel_obj_fun : Callable, optional
            The objective function that takes multiple solutions simultaneously and
            implement the parallelization by itself, by default None.
        eq_fun : Callable, optional
            The equality constraints, whose return value should have the same size as the
            number of equality constraints, by default None.
        ineq_fun : Callable, optional
            The inequality constraints, whose return value should have the same size as
            the number of inequality constraints, by default None.
        model : Any, optional
            The surrogate mode, which will be automatically created if not passed in,
            by default None.
        eval_type : str, optional
            The type of input argument allowed by `obj_func` or `parallel_obj_fun`:
            it could be either 'list' or 'dict', by default 'list'.
        DoE_size : int, optional
            The size of inital Design of Experiment (DoE), by default None.
        warm_data: Tuple, optional
            The warm-starting data in a pair of design points and its objective values
            `(X, y)`, where `X` should be a list of points and has the same length with `y`.
            When provided, the initial sampling (DoE) operation is skipped.
        n_point : int, optional
            The number of candidate solutions proposed using infill-criteria, by default 1
        acquisition_fun : str, optional
            The acquisition function, by default 'EI'
        acquisition_par : dict, optional
            Extra parameters to the acquisition function, by default {}
        acquisition_optimization : dict, optional
            Additional parameters controlling the acquisition optimization, by default {}
        ftarget : float, optional
            The target value to hit, by default None
        max_FEs : int, optional
            The maximal number of evaluations, by default None
        minimize : bool, optional
            To minimize or maximize, by default True
        n_job : int, optional
            The number of allowable jobs for parallelizing the function evaluation
            (if `parallel_obj_fun` is not specified) and Only Effective when n_point > 1,
            by default 1
        data_file : str, optional
            The name of the file to store extra historical information during the run,
            by default None
        verbose : bool, optional
            The verbosity, by default False
        random_seed : int, optional
            The seed for pseudo-random number generators, by default None
        logger : str, optional
            Name of the logger file, by default None, which turns off the logging behaviour
        """
        self.obj_fun = obj_fun
        self.parallel_obj_fun = parallel_obj_fun
        self.h = eq_fun
        self.g = ineq_fun
        self.n_obj: int = 1
        self.n_job: int = max(1, int(n_job))
        self.n_point: int = max(1, int(n_point))
        self.ftarget = ftarget
        self.minimize = minimize
        self.verbose = verbose
        self.data_file = data_file
        self.max_FEs = int(max_FEs) if max_FEs else np.inf
        self.instance_name = instance_name
        self.metric_meta = None

        self.search_space = search_space
        self.DoE_size = DoE_size

        self.acquisition_fun = acquisition_fun
        self._acquisition_par = acquisition_par
        self._acquisition_callbacks = []  # the callback functions executed after
        # every call of `arg_max_acquisition`
        self.model = model
        self.logger = logger
        self.random_seed = random_seed

        self._eval_type = eval_type
        self._set_internal_optimization(**acquisition_optimization)
        self._get_best = np.min if self.minimize else np.max
        self._init_flatfitness_trial = 2
        self._set_aux_vars()
        self.warm_data = warm_data

    @property
    def acquisition_fun(self):
        return self._acquisition_fun

    @acquisition_fun.setter
    def acquisition_fun(self, fun):
        if isinstance(fun, str):
            self._acquisition_fun = fun
        else:
            assert hasattr(fun, "__call__")
        self._acquisition_fun = fun

    @property
    def DoE_size(self):
        return self._DoE_size

    @DoE_size.setter
    def DoE_size(self, DoE_size):
        if DoE_size:
            if isinstance(DoE_size, str):
                self._DoE_size = int(eval(DoE_size))
            elif isinstance(DoE_size, (int, float)):
                self._DoE_size = int(DoE_size)
            else:
                raise ValueError
        else:
            self._DoE_size = int(self.dim * 5)

    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, seed):
        if seed:
            self._random_seed = int(seed)
            if self._random_seed:
                np.random.seed(self._random_seed)

    @property
    def search_space(self):
        return self._search_space

    @search_space.setter
    def search_space(self, search_space):
        self._search_space = search_space
        self.dim = len(self._search_space)
        self.var_names = self._search_space.var_name
        self.r_index = self._search_space.real_id  # indices of continuous variable
        self.i_index = self._search_space.integer_id  # indices of integer variable
        # indices of categorical variable
        self.d_index = self._search_space.categorical_id

        self.param_type = self._search_space.var_type
        self.N_r = len(self.r_index)
        self.N_i = len(self.i_index)
        self.N_d = len(self.d_index)

    @property
    def warm_data(self):
        return self._warm_data

    @warm_data.setter
    def warm_data(self, data):
        assert self.iter_count == 0  # warm data should only be provided in the begining
        if data is None or len(data) == 0:
            self._warm_data = None
        else:
            X, y = data
            assert len(X) == len(y)

            if isinstance(X, Solution):
                assert X.var_name == self.var_names
            else:
                X = self._to_geno(X)

            X.fitness = y
            X.n_eval = 1
            assert all([isinstance(_, float) for _ in X[:, self.r_index].ravel()])
            assert all([isinstance(_, int) for _ in X[:, self.i_index].ravel()])
            assert all([isinstance(_, str) for _ in X[:, self.d_index].ravel()])
            self._warm_data = X
            self.tell(X, y, warm_start=True)

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, logging.Logger):
            self._logger = logger
            self._logger.propagate = False
            return

        # NOTE: logging.getLogger create new instance based on `name`
        # no new instance will be created if the same name is provided
        name = self.instance_name if self.instance_name else str(id(self))
        self._logger = logging.getLogger(f"{self.__class__.__name__} ({name})")
        self._logger.setLevel(logging.DEBUG)
        fmt = LoggerFormatter()

        # create console handler and set level to the vebosity
        SH = list(filter(lambda h: isinstance(h, logging.StreamHandler), self._logger.handlers))
        if self.verbose and len(SH) == 0:
            sh = logging.StreamHandler(sys.stdout)
            sh.setLevel(logging.INFO)
            sh.setFormatter(fmt)
            self._logger.addHandler(sh)

        # create file handler and set level to debug
        # TODO: perhaps also according to the verbosity?
        # TODOL perhaps create a logger class
        FH = list(filter(lambda h: isinstance(h, logging.FileHandler), self._logger.handlers))
        if logger is not None and len(FH) == 0:
            fh = logging.FileHandler(logger)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(fmt)
            self._logger.addHandler(fh)

        if hasattr(self, "logger"):
            self._logger.propagate = False

    def _set_aux_vars(self):
        self.iter_count = 0
        self.eval_count = 0
        self.stop_dict = {}
        self.hist_f = []

        if self._eval_type == "list":
            self._to_pheno = lambda x: x.tolist()
            self._to_geno = lambda x, index=None, n_eval=1, fitness=None: Solution(
                x,
                var_name=self.var_names,
                n_obj=self.n_obj,
                index=index,
                n_eval=n_eval,
                fitness=fitness,
            )
        elif self._eval_type == "dict":
            self._to_pheno = lambda x: x.to_dict()
            self._to_geno = lambda x, index=None: Solution.from_dict(
                x, index=index, n_obj=self.n_obj
            )

    def _set_internal_optimization(self, **kwargs):
        # TODO: turn this into an Option Class
        if "optimizer" in kwargs:
            self._optimizer = kwargs["optimizer"]
        else:
            if isinstance(self.search_space, RealSpace):
                if hasattr(self.model, "gradient") and self.n_obj == 1:
                    self._optimizer = "BFGS"
                else:
                    self._optimizer = "OnePlusOne_Cholesky_CMA"
            else:
                self._optimizer = "MIES"

        if self._optimizer == "BFGS" and (self.h or self.g):
            self._optimizer = "OnePlusOne_Cholesky_CMA"

        # NOTE: `AQ` -> acquisition
        if "max_FEs" in kwargs:
            self.AQ_max_FEs = arg_to_int(kwargs["max_FEs"])
        else:
            self.AQ_max_FEs = default_AQ_max_FEs[self._optimizer](self.dim)

        self.AQ_n_restart = (
            default_AQ_n_restart(self.dim)
            if "n_restart" not in kwargs
            else arg_to_int(kwargs["n_restart"])
        )
        self.AQ_wait_iter = (
            default_AQ_wait_iter if "wait_iter" not in kwargs else arg_to_int(kwargs["wait_iter"])
        )

        # NOTE: `_h` and `_g` are wrappers of `h` and `g`, which always take lists as input
        self._h, self._g = self.h, self.g
        if self._h is not None:
            self._h = func_with_list_arg(self._h, self._eval_type, self.search_space.var_name)
        if self._g is not None:
            self._g = func_with_list_arg(self._g, self._eval_type, self.search_space.var_name)

    def __set_argmax(self, fixed: Dict = None):
        """Set ``self._argmax_restart`` for optimizing the acquisition function"""
        fixed = {} if fixed is None else fixed
        self._argmax_restart = functools.partial(
            argmax_restart,
            search_space=self.search_space.filter(fixed.keys(), invert=True),
            h=partial_argument(self._h, self.search_space.var_name, fixed) if self._h else None,
            g=partial_argument(self._g, self.search_space.var_name, fixed) if self._g else None,
            eval_budget=self.AQ_max_FEs,
            n_restart=self.AQ_n_restart,
            wait_iter=self.AQ_wait_iter,
            optimizer=self._optimizer,
        )

    def _check_params(self):
        # TODO: add more parameter check-ups
        if np.isinf(self.max_FEs):
            raise ValueError("max_FEs cannot be infinite")

        assert hasattr(AcquisitionFunction, self._acquisition_fun)

    def _compare(self, f1: float, f2: float) -> bool:
        """Test if objecctive value f1 is better than f2"""
        return f1 < f2 if self.minimize else f2 > f1

    @property
    def xopt(self):
        if not hasattr(self, "data"):
            return None
        fopt = self._get_best(self.data.fitness)
        self._xopt = self.data[np.where(self.data.fitness == fopt)[0][0]]
        return self._xopt

    def run(self) -> Tuple[List[Solution], dict]:
        while not self.check_stop():
            self.step()
        return self._to_pheno(self.xopt), self.xopt.fitness, self.stop_dict

    def step(self):
        self.logger.info(f"iteration {self.iter_count} starts...")
        X = self.ask()
        # TODO: add exception handling for evaluating the objective function
        func_vals = self.evaluate(X)
        self.tell(X, func_vals)

    @timeit
    def ask(
        self, n_point: int = None, fixed: Dict[str, Union[float, int, str]] = None
    ) -> Union[List[list], List[dict]]:
        """suggest a list of candidate solutions

        Parameters
        ----------
        n_point : int, optional
            the number of candidates to request, by default None
        fixed : Dict[str, Union[float, int, str]], optional
            a dictionary specifies the decision variables fixed and the value to which those
            are fixed, by default None

        Returns
        -------
        Union[List[list], List[dict]]
            the suggested candidates
        """
        if self.model is not None and self.model.is_fitted:
            n_point = self.n_point if n_point is None else n_point
            msg = f"asking {n_point} points:"
            X = self.arg_max_acquisition(n_point=n_point, fixed=fixed)
            X = self.pre_eval_check(X)  # validate the new candidate solutions
            if len(X) < n_point:
                self.logger.warning(
                    f"iteration {self.iter_count}: duplicated solution found "
                    "by optimization! New points is taken from random design."
                )
                N = n_point - len(X)
                # take random samples if the acquisition optimization failed
                X += self.create_DoE(N, fixed=fixed)
        else:  # take the initial DoE
            n_point = self._DoE_size if n_point is None else n_point
            msg = f"asking {n_point} points (using DoE):"
            X = self.create_DoE(n_point, fixed=fixed)

        if len(X) == 0:
            raise AskEmptyError()

        index = np.arange(len(X))
        if hasattr(self, "data"):
            index += len(self.data)

        X = Solution(X, index=index, var_name=self._search_space.var_name)
        self.logger.info(msg)
        for i, _ in enumerate(X):
            self.logger.info(f"#{i + 1} - {self._to_pheno(X[i])}")

        return self._to_pheno(X)

    @timeit
    def tell(
        self,
        X: List[Union[list, dict]],
        func_vals: List[Union[float, list]],
        h_vals: List[Union[float, list]] = None,
        g_vals: List[Union[float, list]] = None,
        index: List[str] = None,
        warm_start: bool = False,
    ):
        """Tell the BO about the function values of proposed candidate solutions

        Parameters
        ----------
        X : List of Lists or Solution
            The candidate solutions which are usually proposed by the `self.ask` function
        func_vals : List/np.ndarray of reals
            The corresponding function values
        """
        # TODO: implement method to handle known, expensive constraints `h_vals` and `g_vals`
        X = self._to_geno(X, index)
        self.logger.info(f"observing {len(X)} points:")
        for i, _ in enumerate(X):
            X[i].fitness = func_vals[i]
            X[i].n_eval += 1
            if not warm_start:
                self.eval_count += 1

            self.logger.info(
                f"#{i + 1} - fitness: {func_vals[i]}, solution: {self._to_pheno(X[i])}"
            )

        X = self.post_eval_check(X)
        self.data = self.data + X if hasattr(self, "data") else X
        self.update_model()

        if self.data_file is not None:
            X.to_csv(self.data_file, header=False, append=True)

        xopt = self.xopt
        self.logger.info(f"fopt: {xopt.fitness}")
        # show the current penalty value if cheap constraints are present
        if self.h is not None or self.g is not None:
            _penalty = dynamic_penalty(xopt.tolist(), 1, self._h, self._g, minimize=self.minimize)
            self.logger.info(f"penalty: {_penalty[0]:.4e}")
        self.logger.info(f"xopt: {self._to_pheno(xopt)}\n")

        if not self.model.is_fitted:
            self._fBest_DoE = copy(xopt.fitness)  # the best point in the DoE
            self._xBest_DoE = copy(self._to_pheno(xopt))

        if not warm_start:
            self.iter_count += 1
            self.hist_f.append(xopt.fitness)

    @timeit
    def evaluate(self, X) -> List[float]:
        """Evaluate the candidate points and update evaluation info in the dataframe"""
        # Parallelization is handled by the objective function itself
        if self.parallel_obj_fun is not None:
            func_vals = self.parallel_obj_fun(X)
        else:
            if self.n_job > 1:  # or by ourselves..
                func_vals = Parallel(n_jobs=self.n_job)(delayed(self.obj_fun)(x) for x in X)
            else:  # or sequential execution
                func_vals = [self.obj_fun(x) for x in X]
        return func_vals

    def recommend(self) -> Solution:
        if self.xopt is None or len(self.xopt) == 0:
            raise RecommendationUnavailableError()
        return self.xopt

    def create_DoE(self, n_point: int, fixed: Dict = None) -> List:
        """get the initial sample points using Design of Experiemnt (DoE) methods

        Parameters
        ----------
        n_point : int
            the number of sample points to draw

        Returns
        -------
        List
            a list of sample points
        """
        fixed = {} if fixed is None else fixed
        search_space = self.search_space.filter(fixed.keys(), invert=True)

        count = 0
        DoE = []
        while n_point:
            # NOTE: random sampling could generate duplicated points again
            # keep sampling until getting enough points
            X = search_space.sample(
                n_point,
                method="LHS" if n_point > 1 else "uniform",
                h=partial_argument(self._h, self.search_space.var_name, fixed)
                if self._h
                else None,
                g=partial_argument(self._g, self.search_space.var_name, fixed)
                if self._g
                else None,
            ).tolist()
            X = fillin_fixed_value(X, fixed, self.search_space)

            if len(X) != 0:
                X = self.search_space.round(X)
                X = self.pre_eval_check(X)
                DoE += X
                n_point -= len(X)

            count += 1
            if count > 3:  # maximally 3 iterations
                break

        return DoE

    @abstractmethod
    def pre_eval_check(self, X: Solution) -> Solution:
        """This function is meant for checking validaty of the solutions prior to the evaluation

        Parameters
        ----------
        X : Solution
            the candidate solutions

        """
        raise NotImplementedError

    def post_eval_check(self, X: Solution) -> Solution:
        _ = np.isnan(X.fitness) | np.isinf(X.fitness)
        if np.any(_):
            self.logger.warning(
                f"{sum(_)} candidate solutions are removed "
                f"due to failed fitness evaluation: \n{str(X[_, :])}"
            )
            X = X[~_, :]
        return X

    def update_model(self):
        # TODO: implement a proper model selection here
        # TODO: in case r2 is really poor, re-fit the model or log-transform `fitness`?
        data = self.data
        fitness = data.fitness

        # TODO: to standardize the response values to prevent numerical overflow that might
        # appear in the MGF-based acquisition function.
        # Standardization should make it easier to specify the GP prior, compared to
        # rescaling values to the unit interval.
        _std = np.std(fitness)
        if len(fitness) > 5 and np.isclose(_std, 0):
            raise FlatFitnessError()

        fitness_ = (
            fitness if np.isclose(_std, 0) else (fitness - np.mean(fitness)) / np.std(fitness)
        )
        self.fmin, self.fmax = np.min(fitness_), np.max(fitness_)
        self.frange = self.fmax - self.fmin

        self.model.fit(data, fitness_)
        fitness_hat = self.model.predict(data)

        r2 = r2_score(fitness_, fitness_hat)
        MAPE = mean_absolute_percentage_error(fitness_, fitness_hat)
        self.logger.info(f"model r2: {r2}, MAPE: {MAPE}")

    @timeit
    def arg_max_acquisition(
        self, n_point: int = None, return_value: bool = False, fixed: Dict = None
    ) -> List[list]:
        """Global Optimization of the acquisition function / Infill criterion

        Returns
        -------
            candidates: tuple of list,
                candidate solution (in list)
            values: tuple,
                criterion value of the candidate solution
        """
        self.logger.debug("acquisition optimziation...")
        n_point = self.n_point if n_point is None else int(n_point)
        return_dx = self._optimizer == "BFGS"
        self.__set_argmax(fixed)

        if n_point > 1:  # multi-point/batch sequential strategy
            candidates, values = self._batch_arg_max_acquisition(
                n_point=n_point, return_dx=return_dx, fixed=fixed
            )
        else:  # single-point strategy
            criteria = self._create_acquisition(par={}, return_dx=return_dx, fixed=fixed)
            candidates, values = self._argmax_restart(criteria, logger=self.logger)
            candidates, values = [candidates], [values]

        candidates = [c for c in candidates if len(c) != 0]
        candidates = fillin_fixed_value(candidates, fixed, self.search_space)
        for callback in self._acquisition_callbacks:
            callback()

        return (candidates, values) if return_value else candidates

    def _create_acquisition(
        self, fun: str = None, par: dict = None, return_dx: bool = False, fixed: Dict = None
    ) -> Callable:
        fun = fun if fun is not None else self._acquisition_fun
        par = par if par is not None else copy(self._acquisition_par)
        par.update({"model": self.model, "minimize": self.minimize})
        criterion = getattr(AcquisitionFunction, fun)(**par)
        return partial_argument(
            func=functools.partial(criterion, return_dx=return_dx),
            var_name=self.search_space.var_name,
            fixed=fixed,
            reduce_output=return_dx,
        )

    def _batch_arg_max_acquisition(self, n_point: int, return_dx: int, fixed: Dict):
        raise NotImplementedError

    def check_stop(self):
        if self.eval_count >= self.max_FEs:
            self.stop_dict["max_FEs"] = self.eval_count

        if self.ftarget is not None and hasattr(self, "xopt"):
            if self._compare(self.xopt.fitness[0], self.ftarget):
                self.stop_dict["ftarget"] = self.xopt.fitness[0]

        return bool(self.stop_dict)

    def save(self, filename: str):
        # creat the folder if not exist
        dirname = os.path.dirname(filename)
        if dirname != "":
            os.makedirs(dirname, exist_ok=True)

        with open(filename, "wb") as f:
            # NOTE: we need to dump `self.data` first. Otherwise, some
            # attributes of it will be lost
            if hasattr(self, "data"):
                self.data = dill.dumps(self.data)

            FHs = list(filter(lambda h: isinstance(h, logging.FileHandler), self.logger.handlers))
            if len(FHs) == 0:
                _logger = None
            else:
                _logger = FHs[0].baseFilename  # Only take the first log file

            logger = self.logger
            self.logger = _logger

            dill.dump(self, f)

            self.logger = logger
            if hasattr(self, "data"):
                self.data = dill.loads(self.data)
            self.logger.info(f"save to file {filename}...")

    @classmethod
    def load(cls, filename):
        with open(filename, "rb") as f:
            obj = dill.load(f)
            if hasattr(obj, "data"):
                obj.data = dill.loads(obj.data)
            obj.logger = getattr(obj, "_logger")
        return obj



import logging
import os
import random
import re
import string
import sys
from copy import copy

import numpy as np

__authors__ = ["Hao Wang"]

# TODO: re-written those functions to C/Cython
def non_dominated_set_2d(y, minimize=True):
    """
    Argument
    --------
    y : numpy 2d array,
        where the each solution occupies a row
    """
    y = np.asarray(y)
    N, _ = y.shape

    if isinstance(minimize, bool):
        minimize = [minimize]

    minimize = np.asarray(minimize).ravel()
    assert len(minimize) == 1 or minimize.shape == (N,)
    y *= (np.asarray([-1] * N) ** minimize).reshape(-1, 1)

    _ = np.argsort(y[:, 0])[::-1]
    y2 = y[_, 1]
    ND = []
    for i in range(N):
        v = y2[i]
        if not any(v <= y2[ND]) or len(ND) == 0:
            ND.append(i)
    return _[ND]


def non_dominated_set_3d(y, minimize=True):
    pass


def fast_non_dominated_sort(fitness):
    fronts = []
    dominated_set = []
    mu = fitness.shape[1]
    n_domination = np.zeros(mu)

    for i in range(mu):
        p = fitness[:, i]
        p_dominated_set = []
        n_p = 0

        for j in range(mu):
            q = fitness[:, j]
            if i != j:
                # TODO: verify this part
                # check the strict domination
                # allow for duplication points on the same front
                if all(p <= q) and not all(p == q):
                    p_dominated_set.append(j)
                elif all(p >= q) and not all(p == q):
                    n_p += 1

        dominated_set.append(p_dominated_set)
        n_domination[i] = n_p

    # create the first front
    fronts.append(np.nonzero(n_domination == 0)[0].tolist())
    n_domination[n_domination == 0] = -1

    i = 0
    while True:
        for p in fronts[i]:
            p_dominated_set = dominated_set[p]
            n_domination[p_dominated_set] -= 1

        _front = np.nonzero(n_domination == 0)[0].tolist()
        n_domination[n_domination == 0] = -1

        if len(_front) == 0:
            break
        fronts.append(_front)
        i += 1

    return fronts


# TODO: implement this as a C procedure
def proportional_selection(perf, N, minimize=True, replacement=True):
    def select(perf):
        perf_min = np.min(perf)
        interval = np.cumsum((perf - perf_min) / (np.sum(perf) - perf_min * len(perf)))
        return np.nonzero(np.random.rand() <= interval)[0][0]

    perf = np.array(perf)
    if minimize:
        perf = -perf
        perf -= np.min(perf)

    if replacement:
        res = [select(perf) for i in range(N)]
    else:
        assert N <= len(perf)
        perf_ = copy(perf)
        idx = list(range(0, len(perf)))
        res = []
        for i in range(N):
            if len(perf_) == 1:
                res.append(idx[0])
            else:
                _ = select(perf_)
                res.append(idx[_])
                perf_ = np.delete(perf_, _)
                del idx[_]
    return res


# TODO: double check this one. It causes the explosion of step-sizes in MIES
def handle_box_constraint(x, lb, ub):
    """This function transforms x to t w.r.t. the low and high
    boundaries lb and ub. It implements the function T^{r}_{[a,b]} as
    described in Rui Li's PhD thesis "Mixed-Integer Evolution Strategies
    for Parameter Optimization and Their Applications to Medical Image
    Analysis" as alorithm 6.

    """
    x = np.asarray(x, dtype="float")
    shape_ori = x.shape
    x = np.atleast_2d(x)
    lb = np.atleast_1d(lb)
    ub = np.atleast_1d(ub)

    transpose = False
    if x.shape[0] != len(lb):
        x = x.T
        transpose = True

    lb, ub = lb.flatten(), ub.flatten()
    lb_index = np.isfinite(lb)
    up_index = np.isfinite(ub)

    valid = np.bitwise_and(lb_index, up_index)

    LB = lb[valid][:, np.newaxis]
    UB = ub[valid][:, np.newaxis]

    y = (x[valid, :] - LB) / (UB - LB)
    I = np.mod(np.floor(y), 2) == 0
    yprime = np.zeros(y.shape)
    yprime[I] = np.abs(y[I] - np.floor(y[I]))
    yprime[~I] = 1.0 - np.abs(y[~I] - np.floor(y[~I]))

    x[valid, :] = LB + (UB - LB) * yprime

    if transpose:
        x = x.T
    return x.reshape(shape_ori)


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class LoggerFormatter(logging.Formatter):
    """TODO: use relative path for %(pathname)s"""

    default_time_format = "%m/%d/%Y %H:%M:%S"
    default_msec_format = "%s,%02d"
    FORMATS = {
        logging.DEBUG: (
            "%(asctime)s - [%(name)s.%(levelname)s] {%(pathname)s:%(lineno)d} -- %(message)s"
        ),
        logging.INFO: "%(asctime)s - [%(name)s.%(levelname)s] -- %(message)s",
        logging.WARNING: "%(asctime)s - [%(name)s.%(levelname)s] {%(name)s} -- %(message)s",
        logging.ERROR: (
            "%(asctime)s - [%(name)s.%(levelname)s] {%(pathname)s:%(lineno)d} -- %(message)s"
        ),
        "DEFAULT": "%(asctime)s - %(levelname)s -- %(message)s",
    }

    def __init__(self, fmt="%(asctime)s - %(levelname)s -- %(message)s"):
        LoggerFormatter.FORMATS["DEFAULT"] = fmt
        super().__init__(fmt=fmt, datefmt=None, style="%")

    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        _fmt = self._style._fmt

        # Replace the original format with one customized by logging level
        self._style._fmt = self.FORMATS.get(record.levelno, self.FORMATS["DEFAULT"])

        # Call the original formatter class to do the grunt work
        fmt = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = _fmt
        return fmt


def get_logger(logger_id: str, logger_file: str = None, console: bool = False) -> logging.Logger:
    # TODO: perhaps also add verbosity level to file asnd console output
    # NOTE: logging.getLogger create new instance based on `name`
    # no new instance will be created if the same name is provided
    logger = logging.getLogger(logger_id)
    logger.setLevel(logging.DEBUG)
    fmt = LoggerFormatter()

    # create console handler and set level to the vebosity
    SH = list(filter(lambda h: isinstance(h, logging.StreamHandler), logger.handlers))
    if console and len(SH) == 0:
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(logging.INFO)
        sh.setFormatter(fmt)
        logger.addHandler(sh)

    # create file handler and set level to debug
    FH = list(filter(lambda h: isinstance(h, logging.FileHandler), logger.handlers))
    if logger_file is not None and logger_file not in [h.baseFilename for h in FH]:
        fh = logging.FileHandler(logger_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(fmt)
        logger.addHandler(fh)

    logger.propagate = False
    return logger


def random_string(k: int = 15):
    return "".join(random.choices(string.ascii_letters + string.digits, k=k))


def expand_replace(s: str):
    m = re.match(r"${.*}", s)
    for _ in m.group():
        s.replace(_, os.path.expandvars(_))
    return s


if __name__ == "__main__":
    # TODO: this goes to unittest
    np.random.seed(1)
    perf = np.random.randn(20)
    print(perf)
    print(proportional_selection(perf, 20, minimize=False, replacement=False))



import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim import BO, OptimizerPipeline, RealSpace
from bayes_optim.acquisition_optim import OnePlusOne_Cholesky_CMA
from bayes_optim.surrogate import GaussianProcess, trend
from deap import benchmarks


class _BO(BO):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._hist_EI = np.zeros(3)

    def ask(self, n_point=None):
        X = super().ask(n_point=n_point)
        if self.model.is_fitted:
            _criter = self._create_acquisition(fun="EI", par={}, return_dx=False)
            self._hist_EI[(self.iter_count - 1) % 3] = np.mean([_criter(x) for x in X])
        return X

    def check_stop(self):
        _delta = self._fBest_DoE - self.fopt
        if (
            self.iter_count > 1
            and np.mean(self._hist_EI[0 : min(3, self.iter_count - 1)]) < 0.01 * _delta
        ):
            self.stop_dict["low-EI"] = np.mean(self._hist_EI)

        if self.eval_count >= (self.max_FEs / 2):
            self.stop_dict["max_FEs"] = self.eval_count

        return super().check_stop()


np.random.seed(42)
dim = 2
max_FEs = 80
obj_fun = lambda x: benchmarks.griewank(x)[0]
lb, ub = -600, 600

search_space = RealSpace([lb, ub]) * dim
mean = trend.constant_trend(dim, beta=None)

# autocorrelation parameters of GPR
thetaL = 1e-10 * (ub - lb) * np.ones(dim)
thetaU = 10 * (ub - lb) * np.ones(dim)
theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

model = GaussianProcess(
    mean=mean,
    corr="squared_exponential",
    theta0=theta0,
    thetaL=thetaL,
    thetaU=thetaU,
    nugget=1e-5,
    noise_estim=False,
    optimizer="BFGS",
    wait_iter=5,
    random_start=5 * dim,
    eval_budget=100 * dim,
)

bo = _BO(
    search_space=search_space,
    obj_fun=obj_fun,
    model=model,
    eval_type="list",
    DoE_size=10,
    n_point=1,
    acquisition_fun="EI",
    verbose=True,
    minimize=True,
)
cma = OnePlusOne_Cholesky_CMA(dim=dim, obj_fun=obj_fun, lb=lb, ub=ub)


def post_BO(BO):
    xopt = BO.xopt
    dim = BO.dim

    H = BO.model.Hessian(xopt)
    g = BO.model.gradient(xopt)[0]

    w, B = np.linalg.eigh(H)
    M = np.diag(1 / np.sqrt(w)).dot(B.T)
    H_inv = B.dot(np.diag(1 / w)).dot(B.T)
    sigma0 = np.linalg.norm(M.dot(g)) / np.sqrt(dim - 0.5)
    if sigma0 == 0:
        sigma0 = 1 / 5

    if np.isnan(sigma0):
        sigma0 = 1 / 5
        H_inv = np.eye(dim)

    kwargs = {
        "x": xopt,
        "fopt": BO.fopt,
        "sigma": sigma0,
        "C": H_inv,
    }
    return kwargs


pipe = OptimizerPipeline(obj_fun=obj_fun, minimize=True, max_FEs=max_FEs, verbose=True)
pipe.add(bo, transfer=post_BO)
pipe.add(cma)
pipe.run()
